import os
import sys
from pathlib import Path

from django.core.management import BaseCommand
from icecream import ic

from endoreg_db.helpers.data_loader import load_data
from endoreg_db.models import PdfType, RawPdfFile


# python manage.py import_report tests/assets/lux-gastro-report.pdf --verbose --start_ollama
# Dynamic import path manipulation to ensure local development version is used
def ensure_local_lx_anonymizer():
    """
    Checks for a local development version of the lx-anonymizer package and adds it to sys.path if available.

    Returns:
        True if the local lx-anonymizer directory was found and added to sys.path; False otherwise.
    """
    script_dir = Path(
        __file__
    ).parent.parent.parent.parent.parent  # /home/admin/dev/lx-annotate/endoreg-db
    local_lx_anonymizer_path = script_dir / "lx-anonymizer"

    if local_lx_anonymizer_path.exists() and local_lx_anonymizer_path.is_dir():
        # Add the directory containing lx_anonymizer to the Python path
        if str(local_lx_anonymizer_path) not in sys.path:
            sys.path.insert(0, str(local_lx_anonymizer_path))
            print(f"Using local lx-anonymizer from: {local_lx_anonymizer_path}")
            return True
    return False


# Try to use local version, fall back to installed version
local_version_available = ensure_local_lx_anonymizer()


# Now import from lx_anonymizer
try:
    from lx_anonymizer.ollama_service import ollama_service
except ImportError:
    print("Could not import init_ollama_service from local or installed lx_anonymizer")
    raise


class Command(BaseCommand):
    """Management Command to import a report file to the database"""

    help = """
        Imports a .pdf file to the database.
        1. Get center by center name from db (default: university_hospital_wuerzburg)
    """

    def add_arguments(self, parser):
        """
        Defines command-line arguments for the import_report management command.

        Adds options for specifying the report file path, center name, report directory, deletion and save behavior, and controls for initializing the Ollama LLM service.
        """
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="Display verbose output for all commands",
        )

        parser.add_argument(
            "file_path",
            type=str,
            help="Path to the Report file to import",
        )
        parser.add_argument(
            "--center_name",
            type=str,
            default="university_hospital_wuerzburg",
            help="Name of the center to associate with the report",
        )

        parser.add_argument(
            "--report_dir_root",
            type=str,
            default="~/test-data/db_report_dir",
            help="Path to the report directory",
        )

        parser.add_argument(
            "--delete_source",
            action="store_true",
            default=False,
            help="Delete the source report file after importing",
        )

        parser.add_argument(
            "--save",
            action="store_true",
            default=False,
            help="Save the report object to the database",
        )

        parser.add_argument(
            "--start_ollama",
            action="store_true",
            default=False,
            help="Start Ollama server for LLM processing",
        )

        parser.add_argument(
            "--ollama_debug",
            action="store_true",
            default=False,
            help="Enable debug mode for Ollama",
        )

        parser.add_argument(
            "--ollama_timeout",
            type=int,
            default=30,
            help="Maximum time to wait for Ollama to start in seconds",
        )

    def handle(self, *args, **options):
        """
        Handles the import of a report report file into the database, with optional LLM service initialization and anonymization.

        This method validates input options, optionally starts the Ollama LLM service, ensures the existence of required files and directories, determines the report type, processes the report for text and metadata extraction, anonymizes content, and saves the resulting data to the database. Provides verbose output and error handling throughout the process.
        """
        # Load initial or prerequisite data for the application.
        # This may include loading default values, configurations, or lookup table data
        # necessary for the import process or other application functionalities.
        try:
            load_data()
            self.stdout.write(self.style.SUCCESS("Successfully loaded initial data."))
        except Exception as e:
            self.stdout.write(self.style.ERROR(f"Failed to load initial data: {e}"))
            # Depending on the criticality of load_data(), you might want to exit or handle differently.
            # For now, we'll just log the error and continue.

        verbose = options["verbose"]
        center_name = options["center_name"]
        report_dir_root = options["report_dir_root"]
        file_path = options["file_path"]
        delete_source = options["delete_source"]
        save = options["save"]
        start_ollama = options["start_ollama"]
        ollama_debug = options["ollama_debug"]
        ollama_timeout = options["ollama_timeout"]

        if not isinstance(delete_source, bool):
            raise ValueError("delete_source must be a boolean")

        self.stdout.write(
            self.style.SUCCESS(f"Starting report import for {file_path}...")
        )

        if local_version_available:
            self.stdout.write(
                self.style.SUCCESS("Using local development version of lx-anonymizer")
            )

        ollama_proc = None  # Track Ollama process if started
        try:
            # Initialize Ollama service if requested
            if start_ollama:
                self.stdout.write(self.style.SUCCESS("Initializing Ollama service..."))
                try:
                    # Set Ollama environment variables
                    os.environ["OLLAMA_MAX_WAIT_TIME"] = str(ollama_timeout)
                    os.environ["OLLAMA_DEBUG"] = "true" if ollama_debug else "false"

                    # Try to find Ollama binary location from env or common paths
                    ollama_bin = os.environ.get("OLLAMA_BIN")
                    if not ollama_bin:
                        # Try common Nix store paths first
                        for path in [
                            "/run/current-system/sw/bin/ollama",
                            "/nix/store/*/bin/ollama",
                        ]:
                            import glob

                            matches = glob.glob(path)
                            if matches:
                                ollama_bin = matches[0]
                                break

                    if ollama_bin:
                        self.stdout.write(
                            self.style.SUCCESS(f"Using Ollama binary at: {ollama_bin}")
                        )
                        os.environ["OLLAMA_BIN"] = ollama_bin

                    # Start Ollama server process if not already running
                    import shutil
                    import subprocess

                    # Check if ollama is already running
                    try:
                        import requests

                        resp = requests.get(
                            "http://127.0.0.1:11434/api/version", timeout=1
                        )
                        if resp.status_code == 200:
                            self.stdout.write(
                                self.style.SUCCESS("Ollama is already running")
                            )
                        else:
                            self.stdout.write(
                                self.style.WARNING(
                                    f"Ollama returned status code {resp.status_code}"
                                )
                            )
                    except requests.exceptions.RequestException:
                        self.stdout.write(
                            self.style.WARNING(
                                "Ollama is not running, attempting to start..."
                            )
                        )
                        # Find ollama binary
                        ollama_path = ollama_bin or shutil.which("ollama")
                        if ollama_path:
                            self.stdout.write(
                                self.style.SUCCESS(
                                    f"Starting Ollama using {ollama_path}"
                                )
                            )
                            # Start ollama serve in background
                            ollama_proc = subprocess.Popen(
                                [ollama_path, "serve"],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                start_new_session=True,
                            )
                            self.stdout.write(
                                self.style.SUCCESS(
                                    "Ollama server started in background"
                                )
                            )
                        else:
                            self.stdout.write(
                                self.style.ERROR("Ollama binary not found in PATH")
                            )

                    # Start the service with explicit initialization
                    ollama_service(auto_start=True)
                    self.stdout.write(
                        self.style.SUCCESS("Ollama service initialized successfully")
                    )
                except Exception as e:
                    self.stdout.write(
                        self.style.ERROR(f"Failed to initialize Ollama service: {e}")
                    )
                    self.stdout.write(
                        self.style.WARNING(
                            "Continuing without Ollama - some features may not work"
                        )
                    )

            # Ensure the report file exists
            file_path = Path(file_path).expanduser()
            if not file_path.exists():
                self.stdout.write(
                    self.style.ERROR(f"Report file not found: {file_path}")
                )
                return

            # Ensure the report directory exists
            report_dir_root = Path(report_dir_root).expanduser()
            report_dir_root.mkdir(parents=True, exist_ok=True)

            # Create the report file object
            self.stdout.write(
                self.style.SUCCESS(f"Creating RawPdfFile object from {file_path}...")
            )
            report_file_obj = RawPdfFile.create_from_file(
                file_path=file_path,
                center_name=center_name,
                delete_source=delete_source,
                save=save,
            )
            if not report_file_obj:
                self.stdout.write(
                    self.style.ERROR("Failed to create RawPdfFile object.")
                )
                return

            report_file_obj.anonymized = False

            # Assign pdfType to the report file object
            if "report" in file_path.name:
                pdf_type_name = "ukw-endoscopy-examination-report-generic"
            elif "histo" in file_path.name:
                pdf_type_name = "ukw-endoscopy-histology-report-generic"
            elif "AW_PA" in file_path.name:
                pdf_type_name = "rkh-endoscopy-histology-report-generic"
            elif "AW" in file_path.name:
                pdf_type_name = "rkh-endoscopy-examination-report-generic"
            else:
                raise ValueError(f"Unknown report type: {file_path.name}")

            self.stdout.write(self.style.SUCCESS(f"Using report type: {pdf_type_name}"))
            try:
                pdf_type = PdfType.objects.get(name=pdf_type_name)
            except PdfType.DoesNotExist:
                self.stdout.write(
                    self.style.ERROR(
                        f"PdfType with name '{pdf_type_name}' does not exist."
                    )
                )
                return
            report_file_obj.pdf_type = pdf_type

            rr_config = report_file_obj.get_report_reader_config()
            pdf_path = report_file_obj.file.path

            # Import at this point to avoid initializing the module too early
            from lx_anonymizer import ReportReader

            self.stdout.write(self.style.SUCCESS("Creating ReportReader..."))
            rr = ReportReader(**rr_config)

            self.stdout.write(self.style.SUCCESS(f"Processing report: {pdf_path}"))
            text, anonymized_text, report_meta = rr.process_report(
                pdf_path, verbose=verbose
            )

            if verbose:
                ic(text, anonymized_text, report_meta)

            self.stdout.write(self.style.SUCCESS("Processing file..."))
            report_file_obj.process_file(
                text, anonymized_text, report_meta, verbose=verbose
            )

            sensitive_meta = report_file_obj.sensitive_meta
            if verbose:
                ic(report_file_obj.sensitive_meta)

            self.stdout.write(self.style.SUCCESS("Saving..."))
            sensitive_meta.save()
            if verbose:
                ic(sensitive_meta)

            report_file_obj.state.anonymization_status.mark_anonymized()
        finally:
            # Clean up Ollama process if we started it
            if ollama_proc is not None:
                import signal

                self.stdout.write(
                    self.style.SUCCESS("Cleaning up Ollama server process...")
                )
                try:
                    ollama_proc.terminate()
                    ollama_proc.wait(timeout=10)
                    self.stdout.write(
                        self.style.SUCCESS("Ollama server process terminated.")
                    )
                except Exception as e:
                    self.stdout.write(
                        self.style.WARNING(
                            f"Failed to terminate Ollama server process: {e}"
                        )
                    )



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Dict, Any, Union

from lx_anonymizer.sensitive_meta_interface import SensitiveMeta
from endoreg_db.models.media import RawPdfFile, VideoFile


@dataclass
class ImportContext:
    """
    Tracking the import success and reasons of failure.
    """

    # core import parameters
    file_path: Path
    center_name: str
    processor_name: str = "olympus-cv-500"
    delete_source: bool = True
    retry: bool = False
    import_completed: bool = False
    error_reason: str = ""

    # paths
    original_path: Optional[Path] = None
    quarantine_path: Optional[Path] = None
    sensitive_path: Optional[Path] = None
    anonymized_path: Optional[Path] = None

    # associated objects
    current_report: Optional[RawPdfFile] = None
    current_video: Optional[VideoFile] = None
    current_meta: Optional[SensitiveMeta] = None
    instance: Optional[Union[RawPdfFile, VideoFile]] = None
    
    file_type: str = "undefined"
    
    
    # processing metadata
    file_hash: str = ""
    original_text: Optional[str] = None
    anonymized_text: Optional[str] = None
    extracted_metadata: Dict[str, Any] = field(default_factory=dict)



"""
Video Processing History Model

Tracks all video correction operations (masking, frame removal, reprocessing).
Created as part of Phase 1.1: Video Correction API Endpoints.
"""

import logging
from pathlib import Path

from django.conf import settings
from django.db import models
from django.utils import timezone

from .video_file import VideoFile

logger = logging.getLogger(__name__)


class VideoProcessingHistory(models.Model):
    """
    History of all video processing operations.

    Stores configuration and results of masking, frame removal, and reprocessing
    operations for audit trail and download access.
    """

    # Operation Types
    OPERATION_MASKING = "mask_overlay"
    OPERATION_FRAME_REMOVAL = "frame_removal"
    OPERATION_ANALYSIS = "analysis"
    OPERATION_REPROCESSING = "reprocessing"

    OPERATION_CHOICES = [
        (OPERATION_MASKING, "Mask Overlay"),
        (OPERATION_FRAME_REMOVAL, "Frame Removal"),
        (OPERATION_ANALYSIS, "Sensitivity Analysis"),
        (OPERATION_REPROCESSING, "Full Reprocessing"),
    ]

    # Status Types
    STATUS_PENDING = "pending"
    STATUS_RUNNING = "running"
    STATUS_SUCCESS = "success"
    STATUS_FAILURE = "failure"
    STATUS_CANCELLED = "cancelled"

    STATUS_CHOICES = [
        (STATUS_PENDING, "Pending"),
        (STATUS_RUNNING, "Running"),
        (STATUS_SUCCESS, "Success"),
        (STATUS_FAILURE, "Failure"),
        (STATUS_CANCELLED, "Cancelled"),
    ]

    video = models.ForeignKey(VideoFile, on_delete=models.CASCADE, related_name="processing_history", help_text="Video file this operation was performed on")

    operation = models.CharField(max_length=50, choices=OPERATION_CHOICES, help_text="Type of processing operation")

    status = models.CharField(max_length=20, choices=STATUS_CHOICES, default=STATUS_PENDING, help_text="Current status of the operation")

    # Configuration & Results
    config = models.JSONField(default=dict, help_text="Operation configuration (mask settings, frame list, etc.)")

    output_file = models.CharField(max_length=500, blank=True, help_text="Path to output file (relative to MEDIA_ROOT)")

    details = models.TextField(blank=True, help_text="Additional details or error messages")

    # Celery Integration
    task_id = models.CharField(max_length=100, blank=True, help_text="Celery task ID for progress tracking")

    # Timestamps
    created_at = models.DateTimeField(auto_now_add=True, help_text="When the operation was started")

    completed_at = models.DateTimeField(null=True, blank=True, help_text="When the operation completed (success or failure)")

    class Meta:
        db_table = "video_processing_history"
        verbose_name = "Video Processing History"
        verbose_name_plural = "Video Processing Histories"
        ordering = ["-created_at"]
        indexes = [
            models.Index(fields=["video", "-created_at"]),
            models.Index(fields=["task_id"]),
            models.Index(fields=["status"]),
        ]

    def __str__(self):
        operation_display = getattr(self, "get_operation_display", None)
        status_display = getattr(self, "get_status_display", None)

        operation = operation_display() if callable(operation_display) else self.operation
        status = status_display() if callable(status_display) else self.status
        return f"{operation} on {self.video.uuid} - {status}"

    def mark_running(self, save=True):
        """Mark operation as running."""
        self.status = self.STATUS_RUNNING
        if save:
            self.save(update_fields=["status"])

    @staticmethod
    def _normalize_output_path(path: str | Path | None) -> str:
        """Return a path relative to MEDIA_ROOT when possible."""
        if not path:
            return ""

        raw_path = Path(path)
        if not raw_path.is_absolute():
            return raw_path.as_posix()
        media_root = Path(settings.MEDIA_ROOT)

        try:
            relative = raw_path.resolve().relative_to(media_root.resolve())
            return relative.as_posix()
        except (ValueError, RuntimeError):
            logger.warning(
                "Storing absolute output path '%s' because it is outside MEDIA_ROOT (%s).",
                raw_path,
                media_root,
            )
            return raw_path.as_posix()

    def mark_success(self, output_file=None, details=None, save=True):
        """Mark operation as successful."""
        self.status = self.STATUS_SUCCESS
        self.completed_at = timezone.now()
        if output_file:
            self.output_file = self._normalize_output_path(output_file)
        if details:
            self.details = details
        if save:
            self.save(update_fields=["status", "completed_at", "output_file", "details"])

    def mark_failure(self, error_message, save=True):
        """Mark operation as failed."""
        self.status = self.STATUS_FAILURE
        self.completed_at = timezone.now()
        self.details = error_message
        if save:
            self.save(update_fields=["status", "completed_at", "details"])

    @property
    def duration(self):
        """Calculate operation duration if completed."""
        if self.completed_at and self.created_at:
            return (self.completed_at - self.created_at).total_seconds()
        return None

    @property
    def is_complete(self) -> bool:
        """Check if operation is in a terminal state."""
        return self.status in [self.STATUS_SUCCESS, self.STATUS_FAILURE, self.STATUS_CANCELLED]



import torch

from .custom_logger import get_logger

logger = get_logger(__name__)

def clear_gpu_memory():
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        logger.debug("Cleared GPU memory.")
        




import os
from pathlib import Path
from icecream import ic


ic(f"LOADING SETTINGS: {__file__}")

DEBUG=True
SECRET_KEY = "fake-key"
INSTALLED_APPS = [
    "tests",
    "endoreg_db.apps.EndoregDbConfig",
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
]
DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

BASE_DIR = Path(__file__).parent
print(f"doc_settings.py - BASE_DIR: {BASE_DIR}")

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': str(BASE_DIR / 'docs_db.sqlite3'),
    },
}

TIME_ZONE = "Europe/Berlin"



# # Path: endoreg_db/forms/questionnaires/tto_questionnaire.py
# from endoreg_db.models import TtoQuestionnaire
# from django import forms

# class TtoQuestionnaireForm(forms.ModelForm):
#     class Meta:
#         model = TtoQuestionnaire
#         fields = '__all__'
#         widgets = {
#             # Specify Bootstrap classes for other fields as needed
#             'patient_name': forms.TextInput(attrs={'class': 'form-control'}),
#             'birth_date': forms.DateInput(attrs={'class': 'form-control', 'type': 'date'}),
#             # You can continue to specify widgets for other fields
#         }


# from django.views.generic.edit import CreateView
# from django.urls import reverse_lazy

# class TtoQuestionnaireCreate(CreateView):
#     model = TtoQuestionnaire
#     form_class = TtoQuestionnaireForm
#     success_url = reverse_lazy('questionnaire_success')  # Redirect to a success page after submission



from rest_framework import serializers
from endoreg_db.models.upload_job import UploadJob


class UploadJobStatusSerializer(serializers.ModelSerializer):
    """
    Read-only serializer for upload job status responses.
    Returns status information for polling endpoints.
    """
    
    sensitive_meta_id = serializers.IntegerField(
        source='sensitive_meta.id',
        read_only=True,
        allow_null=True,
        help_text="ID of the created SensitiveMeta record (only when anonymized)"
    )
    
    # Optional helper fields for preview (can be populated by view if needed)
    text = serializers.CharField(read_only=True, required=False, allow_blank=True)
    anonymized_text = serializers.CharField(read_only=True, required=False, allow_blank=True)

    class Meta:
        model = UploadJob
        fields = [
            'status',
            'error_detail', 
            'sensitive_meta_id',
            'id',
            'text',
            'anonymized_text'
        ]
        read_only_fields = fields

    def to_representation(self, instance):
        """
        Customize the representation to only include relevant fields based on status.
        """
        data = super().to_representation(instance)
        
        # Only include error_detail if status is error
        if instance.status != UploadJob.Status.ERROR:
            data.pop('error_detail', None)
        
        # Only include sensitive_meta_id if status is anonymized and we have a meta record
        if instance.status != UploadJob.Status.ANONYMIZED or not instance.sensitive_meta:
            data.pop('sensitive_meta_id', None)
        
        # Remove empty optional fields
        if not data.get('text'):
            data.pop('text', None)
        if not data.get('anonymized_text'):
            data.pop('anonymized_text', None)
            
        return data


class UploadCreateResponseSerializer(serializers.Serializer):
    """
    Serializer for the initial upload response.
    Returns upload_id and status_url for polling.
    """
    
    upload_id = serializers.UUIDField(
        read_only=True,
        help_text="UUID of the created upload job"
    )
    
    status_url = serializers.CharField(
        read_only=True,
        help_text="URL to poll for upload status updates"
    )


from endoreg_db.models import Examination, Finding


from django.shortcuts import get_object_or_404
from rest_framework.decorators import api_view
from rest_framework.response import Response


@api_view(["GET"])
def get_interventions_for_finding(request, exam_id, finding_id):
    exam = get_object_or_404(Examination, id=exam_id)
    finding = get_object_or_404(Finding, id=finding_id)
    exam_findings = exam.get_available_findings()
    if finding not in exam_findings:
        return Response({"error": "Finding not found for this examination"}, status=404)
    interventions = finding.finding_interventions.all()
    return Response([{"id": i.id, "name": i.name} for i in interventions])


"""
Video processing tasks for Celery-based asynchronous video processing.

This module provides Celery tasks for:
- Video masking with streaming FFmpeg processing
- Frame removal with streaming optimization
- Video reprocessing workflows
- Progress tracking and status updates
"""

from pathlib import Path
from typing import Dict, Any, List, Optional
from celery import shared_task
from celery.utils.log import get_task_logger
from django.shortcuts import get_object_or_404

logger = get_task_logger(__name__)

@shared_task(bind=True)
def apply_video_mask_task(self, video_id: int, mask_type: str = 'device_default', 
                         device_name: str = 'olympus_cv_1500', use_streaming: bool = True,
                         custom_mask: Optional[Dict[str, Any]] = None):
    """
    Apply mask to video using streaming processing.
    
    Args:
        video_id: ID of the VideoFile to process
        mask_type: Type of mask to apply ('device_default', 'roi_based', 'custom')
        device_name: Device name for default mask
        use_streaming: Whether to use streaming processing (recommended)
        custom_mask: Custom mask configuration (if mask_type='custom')
    """
    try:
        # Update task progress
        self.update_state(state='PROGRESS', meta={'progress': 0, 'message': 'Initializing mask application...'})
        
        # Import models here to avoid circular imports
        from endoreg_db.models import VideoFile
        from lx_anonymizer.frame_cleaner import FrameCleaner
        
        # Get video file
        video = get_object_or_404(VideoFile, pk=video_id)
        video_path = Path(video.file.path)
        
        if not video_path.exists():
            raise FileNotFoundError(f"Video file not found: {video_path}")
        
        # Create output path
        output_dir = video_path.parent / "processed"
        output_dir.mkdir(exist_ok=True)
        output_path = output_dir / f"{video_path.stem}_masked{video_path.suffix}"
        
        self.update_state(state='PROGRESS', meta={'progress': 10, 'message': 'Setting up FrameCleaner...'})
        
        # Initialize FrameCleaner
        cleaner = FrameCleaner()
        
        # Determine mask configuration
        if mask_type == 'custom' and custom_mask:
            mask_config = custom_mask
        elif mask_type == 'roi_based':
            # Would need ROI data from video metadata
            mask_config = cleaner._load_mask(device_name)  # Fallback for now
        else:  # device_default
            mask_config = cleaner._load_mask(device_name)
        
        self.update_state(state='PROGRESS', meta={'progress': 20, 'message': f'Applying {mask_type} mask...'})
        
        # Apply mask using streaming
        if use_streaming:
            success = cleaner._mask_video_streaming(
                video_path, mask_config, output_path, use_named_pipe=True
            )
        else:
            success = cleaner._mask_video(video_path, mask_config, output_path)
        
        if not success:
            raise RuntimeError("Video masking failed")
        
        self.update_state(state='PROGRESS', meta={'progress': 90, 'message': 'Finalizing...'})
        
        # Verify output
        if not output_path.exists() or output_path.stat().st_size == 0:
            raise RuntimeError("Output video is empty or missing")
        
        result = {
            'status': 'completed',
            'output_path': str(output_path),
            'mask_type': mask_type,
            'device_name': device_name,
            'use_streaming': use_streaming,
            'output_size': output_path.stat().st_size,
            'message': 'Video masking completed successfully'
        }
        
        logger.info(f"Video masking completed for video {video_id}: {output_path}")
        return result
        
    except Exception as e:
        logger.error(f"Video masking failed for video {video_id}: {e}")
        self.update_state(
            state='FAILURE',
            meta={'error': str(e), 'message': f'Video masking failed: {e}'}
        )
        raise


def _setup_frame_removal(video_id: int, detection_engine: str):
    from endoreg_db.models import VideoFile
    from lx_anonymizer.frame_cleaner import FrameCleaner
    from django.shortcuts import get_object_or_404
    video = get_object_or_404(VideoFile, pk=video_id)
    video_path = Path(video.raw_file.path)
    if not video_path.exists():
        raise FileNotFoundError(f"Video file not found: {video_path}")
    output_dir = video_path.parent / "processed"
    output_dir.mkdir(exist_ok=True)
    output_path = output_dir / f"{video_path.stem}_cleaned{video_path.suffix}"
    use_minicpm = detection_engine == 'minicpm'
    cleaner = FrameCleaner()
    return video, video_path, output_path, cleaner

def _detect_sensitive_frames(self, cleaner, video_path, selection_method, manual_frames, total_frames):
    frames_to_remove = []
    if selection_method == 'manual' and manual_frames:
        frames_to_remove = manual_frames
        self.update_state(state='PROGRESS', meta={'progress': 50, 'message': f'Using manual frame selection: {len(frames_to_remove)} frames'})
    elif selection_method == 'automatic':
        self.update_state(state='PROGRESS', meta={'progress': 20, 'message': 'Detecting sensitive frames...'})
        if total_frames > 10000:
            self.update_state(state='PROGRESS', meta={'progress': 30, 'message': 'Using statistical sampling for large video...'})
            sensitive_frames, estimated_ratio, early_stopped = cleaner._sample_frames_coroutine(
                video_path, total_frames, max_samples=500
            )
            frames_to_remove = sensitive_frames
        else:
            self.update_state(state='PROGRESS', meta={'progress': 30, 'message': 'Analyzing all frames...'})
            for abs_i, gray_frame, skip in cleaner._iter_video(video_path, total_frames):
                progress = 30 + (abs_i / total_frames) * 40  # 30-70% for detection
                if abs_i % 100 == 0:
                    self.update_state(state='PROGRESS', meta={
                        'progress': int(progress),
                        'message': f'Analyzing frame {abs_i}/{total_frames}...'
                    })
                ocr_text, avg_conf, _ = cleaner.frame_ocr.extract_text_from_frame(
                    gray_frame, roi=None, high_quality=True
                )
                if ocr_text:
                    frame_metadata = cleaner.extract_metadata_deepseek(ocr_text)
                    if not frame_metadata:
                        frame_metadata = cleaner.frame_metadata_extractor.extract_metadata_from_frame_text(ocr_text)
                    has_sensitive = cleaner.frame_metadata_extractor.is_sensitive_content(frame_metadata)
                    if has_sensitive:
                        frames_to_remove.append(abs_i)
    return frames_to_remove

def _remove_frames_from_video(cleaner, video_path, frames_to_remove, output_path, use_streaming):
    if use_streaming:
        return cleaner.remove_frames_from_video_streaming(
            video_path, frames_to_remove, output_path, use_named_pipe=True
        )
    else:
        return cleaner.remove_frames_from_video(
            video_path, frames_to_remove, output_path
        )

def _verify_frame_removal_output(output_path):
    if not output_path.exists() or output_path.stat().st_size == 0:
        raise RuntimeError("Output video is empty or missing")

@shared_task(bind=True)
def remove_video_frames_task(self, video_id: int, selection_method: str = 'automatic',
                            detection_engine: str = 'minicpm', use_streaming: bool = True,
                            manual_frames: Optional[List[int]] = None):
    """
    Remove frames from video using streaming processing.
    
    Args:
        video_id: ID of the VideoFile to process
        selection_method: 'automatic' or 'manual'
        detection_engine: 'minicpm' or 'traditional'
        use_streaming: Whether to use streaming processing
        manual_frames: List of frame indices to remove (for manual selection)
    """
    try:
        # Update task progress
        self.update_state(state='PROGRESS', meta={'progress': 0, 'message': 'Initializing frame removal...'})
        
        video, video_path, output_path, cleaner = _setup_frame_removal(video_id, detection_engine)
        
        self.update_state(state='PROGRESS', meta={'progress': 10, 'message': 'Setting up FrameCleaner...'})
        
        # Get total frame count
        total_frames = cleaner._get_total_frames()
        
        frames_to_remove = _detect_sensitive_frames(
            self, cleaner, video_path, selection_method, manual_frames, total_frames
        )
        
        self.update_state(state='PROGRESS', meta={'progress': 70, 'message': f'Removing {len(frames_to_remove)} frames...'})
        
        # Remove frames using streaming
        success = _remove_frames_from_video(cleaner, video_path, frames_to_remove, output_path, use_streaming)

        if not success:
            raise RuntimeError("Frame removal failed")
        
        self.update_state(state='PROGRESS', meta={'progress': 90, 'message': 'Finalizing...'})
        
        _verify_frame_removal_output(output_path)
        
        result = {
            'status': 'completed',
            'output_path': str(output_path),
            'frames_removed': len(frames_to_remove),
            'selection_method': selection_method,
            'detection_engine': detection_engine,
            'use_streaming': use_streaming,
            'output_size': output_path.stat().st_size,
            'message': f'Successfully removed {len(frames_to_remove)} frames'
        }
        
        logger.info(f"Frame removal completed for video {video_id}: {len(frames_to_remove)} frames removed")
        return result
        
    except Exception as e:
        logger.error(f"Frame removal failed for video {video_id}: {e}")
        self.update_state(
            state='FAILURE',
            meta={'error': str(e), 'message': f'Frame removal failed: {e}'}
        )
        raise


@shared_task(bind=True)
def reprocess_video_task(self, video_id: int):
    """
    Reprocess video with updated settings.
    
    Args:
        video_id: ID of the VideoFile to reprocess
    """
    try:
        self.update_state(state='PROGRESS', meta={'progress': 0, 'message': 'Starting video reprocessing...'})
        
        # Import models here to avoid circular imports
        from endoreg_db.models import VideoFile
        from lx_anonymizer.frame_cleaner import FrameCleaner
        
        # Get video file
        video = get_object_or_404(VideoFile, pk=video_id)
        video_path = Path(video.file.path)
        
        if not video_path.exists():
            raise FileNotFoundError(f"Video file not found: {video_path}")
        
        self.update_state(state='PROGRESS', meta={'progress': 20, 'message': 'Initializing FrameCleaner...'})
        
        # Initialize FrameCleaner with optimal settings
        cleaner = FrameCleaner()
        
        # Create output path
        output_dir = video_path.parent / "processed"
        output_dir.mkdir(exist_ok=True)
        output_path = output_dir / f"{video_path.stem}_reprocessed{video_path.suffix}"
        
        self.update_state(state='PROGRESS', meta={'progress': 30, 'message': 'Analyzing video...'})
        
        # Perform full analysis
        analysis_result = cleaner.analyze_video_sensitivity()
        
        self.update_state(state='PROGRESS', meta={'progress': 60, 'message': f'Processing with {analysis_result["recommended_method"]}...'})
        
        # Process based on analysis recommendation
        if analysis_result["recommended_method"] == "masking":
            # Apply device-specific masking
            mask_config = cleaner._load_mask("olympus_cv_1500")  # Default device
            success = cleaner._mask_video_streaming(
                video_path, mask_config, output_path, use_named_pipe=True
            )
        else:
            # Remove sensitive frames
            sensitive_frames = analysis_result.get("sensitive_frame_list", [])
            # Filter out string elements like "...truncated"
            sensitive_frames = [f for f in sensitive_frames if isinstance(f, int)]
            
            success = cleaner.remove_frames_from_video_streaming(
                video_path, sensitive_frames, output_path, use_named_pipe=True
            )
        
        if not success:
            raise RuntimeError("Video reprocessing failed")
        
        self.update_state(state='PROGRESS', meta={'progress': 90, 'message': 'Finalizing...'})
        
        # Update video metadata with analysis results
        if hasattr(video, 'sensitive_frame_count'):
            video.sensitive_frame_count = analysis_result.get('sensitive_frames', 0)
        if hasattr(video, 'total_frames'):
            video.total_frames = analysis_result.get('total_frames', 0)
        if hasattr(video, 'sensitive_ratio'):
            video.sensitive_ratio = analysis_result.get('sensitivity_ratio', 0.0)
        
        try:
            video.save()
        except Exception as e:
            logger.warning(f"Could not save video metadata: {e}")
        
        result = {
            'status': 'completed',
            'output_path': str(output_path),
            'analysis_result': analysis_result,
            'processing_method': analysis_result["recommended_method"],
            'output_size': output_path.stat().st_size if output_path.exists() else 0,
            'message': 'Video reprocessing completed successfully'
        }
        
        logger.info(f"Video reprocessing completed for video {video_id}")
        return result
        
    except Exception as e:
        logger.error(f"Video reprocessing failed for video {video_id}: {e}")
        self.update_state(
            state='FAILURE',
            meta={'error': str(e), 'message': f'Video reprocessing failed: {e}'}
        )
        raise


import logging
import shutil
import uuid
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Type

# Import the new exceptions from the correct path
from endoreg_db.exceptions import InsufficientStorageError, TranscodingError
from endoreg_db.utils.paths import IMPORT_VIDEO_DIR, SENSITIVE_VIDEO_DIR, TRANSCODING_DIR

if TYPE_CHECKING:
    from endoreg_db.models import VideoFile

import endoreg_db.utils.paths as path_utils
from ....utils.file_operations import get_uuid_filename
from ....utils.hashs import get_video_hash
from ....utils.video.ffmpeg_wrapper import transcode_videofile_if_required

logger = logging.getLogger(__name__)


def check_storage_capacity(
    src_path: Path, dst_root: Path, safety_margin: float = 1.2
) -> None:
    """
    Check if there's enough storage space before starting operations.

    Args:
        src_path: Source file path
        dst_root: Destination root directory
        safety_margin: Safety factor (1.2 = 20% extra space required)

    Raises:
        InsufficientStorageError: If insufficient storage space
    """
    try:
        src_size = src_path.stat().st_size
        required_space = int(src_size * safety_margin)

        # Check free space on destination
        free_space = shutil.disk_usage(dst_root).free

        if free_space < required_space:
            raise InsufficientStorageError(
                f"Insufficient storage space. Required: {required_space / 1e9:.1f} GB, Available: {free_space / 1e9:.1f} GB on {dst_root}",
                required_space=required_space,
                available_space=free_space,
            )

        logger.info(
            f"Storage check passed: {free_space / 1e9:.1f} GB available, {required_space / 1e9:.1f} GB required"
        )

    except OSError as e:
        logger.warning(f"Could not check storage capacity: {e}")
        # Don't fail the operation, just log the warning


def atomic_copy_with_fallback(src_path: Path = IMPORT_VIDEO_DIR, dst_path: Path = SENSITIVE_VIDEO_DIR) -> bool:
    """
    Atomically copy file from src to dst, preserving the source file.

    Args:
        src_path: Source file path
        dst_path: Destination file path

    Returns:
        True if successful

    Raises:
        InsufficientStorageError: If not enough space for the operation
        OSError: For other file system errors
    """
    try:
        # Check space before copy
        src_size = src_path.stat().st_size
        free_space = shutil.disk_usage(dst_path.parent).free

        if free_space < src_size * 1.1:  # 10% safety margin
            raise InsufficientStorageError(
                f"Insufficient space for copy operation. Required: {src_size / 1e9:.1f} GB, Available: {free_space / 1e9:.1f} GB",
                required_space=src_size,
                available_space=free_space,
            )

        # Use a temporary name during copy for atomicity
        temp_dst = dst_path.with_suffix(dst_path.suffix + ".tmp")

        try:
            shutil.copy2(str(src_path), str(temp_dst))
            temp_dst.rename(dst_path)
            logger.debug(f"Copy successful: {src_path} -> {dst_path}")
            return True
        except Exception:
            # Clean up temp file if copy failed
            if temp_dst.exists():
                temp_dst.unlink(missing_ok=True)
            raise

    except Exception as e:
        logger.error(f"Copy operation failed: {src_path} -> {dst_path}: {e}")
        raise


def atomic_move_with_fallback(src_path: Path, dst_path: Path) -> bool:
    """
    Atomically move file from src to dst, with fallback to copy+remove.

    Args:
        src_path: Source file path
        dst_path: Destination file path

    Returns:
        True if successful

    Raises:
        InsufficientStorageError: If not enough space for the operation
        OSError: For other file system errors
    """
    try:
        # First try atomic move (same filesystem)
        try:
            src_path.rename(dst_path)
            logger.debug(f"Atomic move successful: {src_path} -> {dst_path}")
            return True
        except OSError as e:
            if e.errno == 18:  # Cross-device link
                logger.debug("Cross-device move detected, falling back to copy+remove")
            else:
                raise

        # Check space before cross-filesystem copy
        src_size = src_path.stat().st_size
        free_space = shutil.disk_usage(dst_path.parent).free

        if free_space < src_size * 1.1:  # 10% safety margin
            raise InsufficientStorageError(
                f"Insufficient space for copy operation. Required: {src_size / 1e9:.1f} GB, Available: {free_space / 1e9:.1f} GB",
                required_space=src_size,
                available_space=free_space,
            )

        # Fallback to copy+remove for cross-filesystem moves
        logger.info(f"Copying file (cross-filesystem): {src_path} -> {dst_path}")

        # Use a temporary name during copy for atomicity
        temp_dst = dst_path.with_suffix(dst_path.suffix + ".tmp")

        try:
            shutil.copy2(str(src_path), str(temp_dst))
            temp_dst.rename(dst_path)
            src_path.unlink()  # Remove source only after successful copy
            logger.debug(f"Copy+remove successful: {src_path} -> {dst_path}")
            return True

        except OSError as e:
            # Clean up temp file on failure
            if temp_dst.exists():
                temp_dst.unlink(missing_ok=True)
            # Re-raise with better context
            if e.errno == 28:  # No space left on device
                raise InsufficientStorageError(
                    f"No space left on device during copy: {e}",
                    required_space=src_path.stat().st_size,
                    available_space=shutil.disk_usage(dst_path.parent).free,
                )
            raise

    except Exception as e:
        logger.error(f"Failed to move {src_path} -> {dst_path}: {e}")
        raise


def _get_data_paths():
    """Return the current data_paths mapping (supports patched instances in tests)."""
    utils_module = import_module("endoreg_db.utils")
    return getattr(utils_module, "data_paths")


def _get_path(mapping, key, default):
    """Access mapping by key using __getitem__ so MagicMocks with side effects work."""
    if mapping is None:
        return default
    try:
        return mapping[key]
    except (KeyError, TypeError):
        return default


def _create_from_file(
    cls_model: Type["VideoFile"],
    file_path: Path,
    center_name: str,
    processor_name: Optional[str] = None,
    video_dir: Path = IMPORT_VIDEO_DIR,
    save: bool = True,
    delete_source: bool = False,
    **kwargs,
) -> "VideoFile":
    """
    Creates a VideoFile instance from a given video file path with improved error handling.

    Raises:
        InsufficientStorageError: When not enough disk space
        TranscodingError: When video transcoding fails
        ValueError: When required objects (Center, Processor) not found
        RuntimeError: For other processing errors
    """
    from endoreg_db.models.administration.center.center import Center
    from endoreg_db.models.medical.hardware import EndoscopyProcessor

    original_file_name = file_path.name
    original_suffix = file_path.suffix
    final_storage_path = None
    transcoded_file_path = None

    try:
        # Ensure we operate under the canonical video path root
        data_paths = _get_data_paths()
        resolved_video_dir = _get_path(data_paths, "video", video_dir)
        video_dir = Path(resolved_video_dir)
        storage_root_default = Path(video_dir).parent
        resolved_storage_root = _get_path(data_paths, "storage", storage_root_default)
        storage_root = Path(resolved_storage_root)
        storage_root.mkdir(parents=True, exist_ok=True)

        # Check storage capacity before starting any work
        check_storage_capacity(file_path, storage_root)

        # 1. Transcode if necessary
        logger.debug("Checking transcoding requirement for %s", file_path)
        temp_transcode_dir = TRANSCODING_DIR
        temp_transcode_dir.mkdir(parents=True, exist_ok=True)

        # Use a unique name for the potential transcoded file
        temp_transcoded_output_path = (
            temp_transcode_dir / f"{uuid.uuid4()}{original_suffix}"
        )

        try:
            transcoded_file_path = transcode_videofile_if_required(
                input_path=file_path, output_path=temp_transcoded_output_path
            )
            if transcoded_file_path is None:
                raise TranscodingError(
                    f"Transcoding check/process failed for {file_path}"
                )
        except Exception as e:
            raise TranscodingError(f"Video transcoding failed: {e}") from e

        logger.debug("Using file for hashing: %s", transcoded_file_path)

        # 2. Calculate hash (this will be the raw_video_hash)
        video_hash = get_video_hash(transcoded_file_path)
        if not video_hash:
            raise ValueError(
                f"Could not calculate video hash for {transcoded_file_path}"
            )
        logger.info(
            "Calculated raw video hash: %s for %s", video_hash, original_file_name
        )

        # 3. Check if hash already exists
        if cls_model.check_hash_exists(video_hash=video_hash):
            existing_video = cls_model.objects.get(video_hash=video_hash)
            logger.warning(
                "Video with hash %s already exists (UUID: %s)",
                video_hash,
                existing_video.uuid,
            )

            # Check if the existing video has a valid file
            existing_raw_path = existing_video.get_raw_file_path()
            if (
                existing_video.has_raw
                and existing_raw_path
                and existing_raw_path.exists()
            ):
                logger.warning(
                    "Video with hash %s already exists and file is present. Returning existing instance.",
                    video_hash,
                )
                # Clean up transcoded file if it was created temporarily
                if transcoded_file_path != file_path and transcoded_file_path.exists():
                    transcoded_file_path.unlink(missing_ok=True)
                return existing_video

            logger.warning(
                "Video with hash %s exists but file is missing. Deleting orphaned record.",
                video_hash,
            )
            existing_video.delete()

        # 4. Generate UUID and final storage path
        new_file_name, uuid_val = get_uuid_filename(transcoded_file_path)
        final_storage_path = video_dir / new_file_name
        final_storage_path.parent.mkdir(parents=True, exist_ok=True)

        # 5. Move or Copy the file to final storage using improved method
        try:
            if delete_source and transcoded_file_path == file_path:
                logger.debug(
                    "Moving original file %s to %s", file_path, final_storage_path
                )
                atomic_move_with_fallback(file_path, final_storage_path)
            elif delete_source and transcoded_file_path != file_path:
                logger.debug(
                    "Moving transcoded file %s to %s",
                    transcoded_file_path,
                    final_storage_path,
                )
                atomic_move_with_fallback(transcoded_file_path, final_storage_path)
            else:
                logger.debug(
                    "Copying file %s to %s", transcoded_file_path, final_storage_path
                )
                atomic_copy_with_fallback(transcoded_file_path, final_storage_path)
                if transcoded_file_path != file_path and transcoded_file_path.exists():
                    logger.debug(
                        "Cleaning up temporary transcoded file %s", transcoded_file_path
                    )
                    transcoded_file_path.unlink(missing_ok=True)
        except InsufficientStorageError:
            # Re-raise storage errors as-is
            raise
        except Exception as e:
            raise RuntimeError(f"Failed to move file to final storage: {e}") from e

        # 6. Verify hash after move/copy
        final_hash = get_video_hash(final_storage_path)
        if final_hash != video_hash:
            logger.error(
                "Hash mismatch after file operation! Expected %s, got %s",
                video_hash,
                final_hash,
            )
            final_storage_path.unlink(missing_ok=True)
            raise RuntimeError(
                f"Hash mismatch after file operation for {final_storage_path}"
            )

        # 7. Get related objects
        try:
            center = Center.objects.get(name=center_name)
            processor = (
                EndoscopyProcessor.objects.get(name=processor_name)
                if processor_name
                else None
            )
            logger.debug(
                "Found Center: %s, Processor: %s",
                center.name,
                processor.name if processor else "None",
            )
        except Center.DoesNotExist as e:
            logger.error("Center '%s' not found", center_name)
            if final_storage_path and final_storage_path.exists():
                final_storage_path.unlink(missing_ok=True)
            raise ValueError(f"Center '{center_name}' not found.") from e
        except EndoscopyProcessor.DoesNotExist as e:
            logger.error("Processor '%s' not found", processor_name)
            if final_storage_path and final_storage_path.exists():
                final_storage_path.unlink(missing_ok=True)
            raise ValueError(f"Processor '{processor_name}' not found.") from e

        # 8. Create the VideoFile instance
        logger.info("Creating new VideoFile instance with UUID: %s", uuid_val)
        # Store FileField path relative to storage root including the videos prefix
        # TODO Review removal, since this is now newly handled by path_utils
        #storage_base = Path(_get_path(data_paths, "storage", final_storage_path.parent))
        
        relative_name = path_utils.to_storage_relative(final_storage_path)        
        video = cls_model(
            uuid=uuid_val,
            raw_file=relative_name,
            processed_file=None,
            center=center,
            processor=processor,
            original_file_name=original_file_name,
            video_hash=video_hash,
            processed_video_hash=None,
            suffix=original_suffix,
        )

        # 9. Save the instance if requested
        if save:
            logger.info("Saving new VideoFile instance (UUID: %s)", uuid_val)
            video.save()
            logger.info(
                "Successfully created VideoFile PK %s (UUID: %s)", video.pk, video.uuid
            )

        return video

    except (InsufficientStorageError, TranscodingError, ValueError):
        # Re-raise these specific errors as-is
        raise
    except Exception as e:
        logger.error(
            "Failed to create VideoFile from %s: (%s) %s",
            file_path,
            type(e).__name__,
            e,
            exc_info=True,
        )
        # Clean up any created files
        if final_storage_path and final_storage_path.exists():
            logger.warning("Cleaning up orphaned file: %s", final_storage_path)
            final_storage_path.unlink(missing_ok=True)
        if (
            transcoded_file_path
            and transcoded_file_path != file_path
            and transcoded_file_path.exists()
        ):
            logger.warning(
                "Cleaning up orphaned transcoded file: %s", transcoded_file_path
            )
            transcoded_file_path.unlink(missing_ok=True)
        raise RuntimeError(f"Video processing failed: {e}") from e



# endoreg_db/import_files/storage/create_report_file.py
import logging
from typing import Tuple

from endoreg_db.import_files.context.import_context import ImportContext
from endoreg_db.models.media import RawPdfFile
from endoreg_db.models.media.storage.processing_history import ProcessingHistory
from endoreg_db.import_files.context.ensure_center import ensure_center
logger = logging.getLogger(__name__)


def create_or_retrieve_report_file(
    ctx: ImportContext,
) -> Tuple[RawPdfFile, bool]:
    """
    Create a new or retrieve an existing RawPdfFile for the given context.

    Returns:
        pdf             : RawPdfFile instance
        needs_processing: True if the pipeline should run for this file
                          (no successful history yet for this object/file_type key)
    """
    file_path = ctx.file_path
    center_name = ctx.center_name
    delete_source = ctx.delete_source
    file_type = ctx.file_type  # logical key for history; can be None

    # 1) Determine the RawPdfFile instance to work with
    if ctx.current_report is not None:
        pdf = ctx.current_report
        logger.info("Using existing RawPdfFile from context: pk=%s", pdf.pk)
    else:
        logger.info("Creating new RawPdfFile from %s for center %s", file_path, center_name)
        
        pdf = RawPdfFile.create_from_file_initialized(
            file_path=file_path,
            center_name=center_name,
            delete_source=delete_source,
        )
        
        center = ensure_center(pdf, ctx.center_name)
        
        logger.info(f"Successfully set up report file from {center.name}")



    # 3) Check if we already have a successful history entry for this object+file_type
    has_success_history = ProcessingHistory.has_history_for_object(
        obj=pdf,
        success=True,
    )

    if has_success_history:
        logger.info(
            "RawPdfFile %s already has successful processing history (file_type=%s) - short-circuiting",
            getattr(pdf, str(pdf.file_path)),
            file_type,
        )
        # No need to run the pipeline again
        return pdf, False

    # 4) No successful history yet → ensure there is a history entry marking it as "in progress"/failed
    ProcessingHistory.get_or_create_for_object(
        obj=pdf,
        # At this point we haven't finished anonymization; treat as not-success yet.
        success=False,
    )

    logger.info(
        "Report instance ready for processing: pk=%s, file_type=%s (needs_processing=True)",
        pdf.pk,
        file_type,
    )

    # Signal to the caller that the anonymization pipeline should run
    return pdf, True



import logging
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from ..video_file import VideoFile

logger = logging.getLogger(__name__)


def _get_endo_roi(video: "VideoFile") -> Optional[Dict[str, int]]:
    """
    Gets the endoscope region of interest (ROI) dictionary from the linked VideoMeta.

    The ROI dictionary typically contains 'x', 'y', 'width', 'height'.
    Returns None if VideoMeta is not linked or ROI is not properly defined.
    """
    if not video.video_meta:
        logger.warning("VideoMeta not linked for video %s. Cannot get endo ROI.", video.uuid)
        return None

    try:
        # Assuming VideoMeta has a method get_endo_roi()
        endo_roi = video.video_meta.get_endo_roi()
        # Basic validation
        if (
            isinstance(endo_roi, dict)
            and all(k in endo_roi for k in ("x", "y", "width", "height"))
            and all(isinstance(v, int) and not isinstance(v, bool) for v in endo_roi.values())
        ):
            cleaned_roi = {k: int(endo_roi[k] or 0) for k in ("x", "y", "width", "height")}
            logger.debug("Retrieved endo ROI for video %s: %s", video.uuid, cleaned_roi)
            return cleaned_roi
        else:
            logger.warning("Endo ROI not fully defined or invalid in VideoMeta for video %s. ROI: %s", video.uuid, endo_roi)
            return None
    except AttributeError:
        logger.error("VideoMeta object for video %s does not have a 'get_endo_roi' method.", video.uuid)
        return None
    except Exception as e:
        logger.error("Error getting endo ROI from VideoMeta for video %s: %s", video.uuid, e, exc_info=True)
        return None



from typing import TYPE_CHECKING, cast

from django.db import models


class FindingClassificationTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class FindingClassificationType(models.Model):
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(blank=True)
    objects = FindingClassificationTypeManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return str(self.name)


class FindingClassificationManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class FindingClassification(models.Model):
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(blank=True)
    finding_types = models.ManyToManyField("FindingType", blank=True, related_name="finding_classifications")
    choices = models.ManyToManyField("FindingClassificationChoice", related_name="classifications", blank=True)

    classification_types = models.ManyToManyField(
        to=FindingClassificationType,
        # on_delete=models.CASCADE
    )
    information_sources = models.ManyToManyField(
        "InformationSource",
        related_name="finding_classifications",
        blank=True,
    )

    @property
    def examinations(self):
        from endoreg_db.models import Examination

        return Examination.objects.filter(findings__finding_classifications=self)

    objects = FindingClassificationManager()

    if TYPE_CHECKING:
        from endoreg_db.models import Examination, Finding, FindingType, InformationSource, PatientFindingClassification

        classification_types = cast(models.manager.RelatedManager["FindingClassificationType"], classification_types)
        choices = cast(models.manager.RelatedManager["FindingClassificationChoice"], choices)
        finding_types = cast(models.manager.RelatedManager["FindingType"], finding_types)
        information_sources = cast(models.manager.RelatedManager["InformationSource"], information_sources)

        @property
        def findings(self) -> "models.manager.RelatedManager[Finding]": ...

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return str(self.name)

    def get_choices(self):
        """
        Return all choices associated with this classification.

        Returns:
                QuerySet: All related FindingClassificationChoice instances.
        """
        return self.choices.all()


class FindingClassificationChoiceManager(models.Manager):
    def get_by_natural_key(self, name):
        """
        Retrieve an instance by its unique name using the natural key.

        Parameters:
            name (str): The unique name identifying the instance.

        Returns:
            The model instance with the specified name.
        """
        return self.get(name=name)


class FindingClassificationChoice(models.Model):
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(blank=True)
    subcategories = models.JSONField(default=dict)
    numerical_descriptors = models.JSONField(default=dict)
    objects = FindingClassificationChoiceManager()

    if TYPE_CHECKING:
        from endoreg_db.models import PatientFindingClassification

        classifications: models.QuerySet["FindingClassification"]
        patient_finding_classifications: models.QuerySet["PatientFindingClassification"]

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        classifications_names = ", ".join([c.name for c in self.classifications.all()])
        _str = f"{self.name} ({classifications_names})"
        return _str



from django.core.management.base import BaseCommand

from ...data import QUALIFICATION_DATA_DIR
from ...models import Qualification, QualificationType
from ...utils import load_model_data_from_yaml

SOURCE_DIR = QUALIFICATION_DATA_DIR  # qualification data directory

model_0 = QualificationType
model_1 = Qualification

IMPORT_MODELS = [  # string as model key, serves as key in IMPORT_METADATA
    model_0.__name__,
    model_1.__name__,
]

IMPORT_METADATA = {
    model_0.__name__: {
        "dir": SOURCE_DIR,  # e.g. "interventions"
        "model": model_0,
        "foreign_keys": [],  # e.g. ["intervention_types"]
        "foreign_key_models": [],  # e.g. [InterventionType]
    },
    model_1.__name__: {
        "dir": SOURCE_DIR,  # e.g. "interventions"
        "model": model_1,
        "foreign_keys": ["qualification_types"],  # e.g. ["intervention_types"]
        "foreign_key_models": [model_0],  # e.g. [InterventionType]
    },
}


class Command(BaseCommand):
    help = """Load all .yaml files in the data/qualification directory
    into the Qualification / Qualification Type model"""

    def add_arguments(self, parser):
        """
        Adds the --verbose flag to the command-line parser to enable detailed output.
        """
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="Display verbose output",
        )

    def handle(self, *args, **options):
        """
        Loads data from YAML files into the QualificationType and Qualification models.

        Iterates over predefined models and imports their data from YAML files using associated metadata. Supports verbose output if enabled via command-line options.
        """
        verbose = options["verbose"]
        for model_name in IMPORT_MODELS:
            _metadata = IMPORT_METADATA[model_name]
            load_model_data_from_yaml(self, model_name, _metadata, verbose)



# storage/__init__.py

from . import create_report_file
from . import create_video_file
from . import sensitive_meta_storage
from . import state_management
from . import storage

__all__ = [
    "create_report_file",
    "create_video_file",
    "sensitive_meta_storage",
    "state_management",
    "storage",
]



from decimal import Decimal
from typing import TYPE_CHECKING, cast

from django.db import models

if TYPE_CHECKING:
    from endoreg_db.models import (
        Center,
        ScheduledDays,
        Shift,
    )


class CenterShiftManager(models.Manager):
    def get_by_natural_key(self, name):
        """
        Retrieves a CenterShift instance by its unique name.

        Args:
            name: The unique identifier of the CenterShift.

        Returns:
            The CenterShift instance with the specified name.
        """
        return self.get(name=name)


class CenterShift(models.Model):
    """
    Model representing a center shift.
    """

    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(blank=True, null=True)

    center = models.ForeignKey(
        "Center",
        on_delete=models.CASCADE,
        related_name="center_shifts",
    )
    shift = models.ForeignKey(
        "Shift",
        on_delete=models.CASCADE,
        related_name="center_shifts",
    )

    start_time = models.TimeField()
    end_time = models.TimeField()
    scheduled_days = models.ManyToManyField(
        "ScheduledDays",
        related_name="center_shifts",
    )

    # TODO add validator; the value should be between 0 and 1
    estimated_presence_fraction = models.DecimalField(
        max_digits=5,
        decimal_places=4,
        default=Decimal("0"),
    )

    if TYPE_CHECKING:
        center: models.ForeignKey["Center"]
        shift: models.ForeignKey["Shift"]
        scheduled_days = cast(models.manager.RelatedManager["ScheduledDays"], scheduled_days)

    def __str__(self):
        """
        Returns a string representation of the center shift, combining the center and shift.
        """
        return f"{self.center} - {self.shift}"

    def initialize_scheduled_days(self):
        # shift_types
        """
        Initializes scheduled days for the center shift.

        This method is a placeholder and does not perform any actions.
        """
        pass

    def get_scheduled_days(self, infer=True):
        """
        Retrieves the scheduled days associated with this center shift.

        If no scheduled days are found and `infer` is False, raises a ValueError. Otherwise, returns the queryset of scheduled days, which may be empty.
        """
        sd = self.scheduled_days.all()
        if not sd.exists():
            if not infer:
                raise ValueError("No scheduled days found for this center shift.")

        return sd



from django.core.management.utils import get_random_secret_key
import os
import json
from pathlib import Path
import shutil


# --- Constants ---
DEFAULT_DB_PASSWORD = "changeme_in_production" # Placeholder password

# --- Load Nix Variables ---
nix_vars = {}
nix_vars_path = Path("./.devenv-vars.json")
if nix_vars_path.exists():
    with open(nix_vars_path, 'r', encoding="utf-8") as f:
        nix_vars = json.load(f)
    print(f"Loaded Nix variables: {', '.join(nix_vars.keys())}")
else:
    print("No Nix variables file found at .devenv-vars.json")

# --- Determine Paths ---
# Use WORKING_DIR from Nix vars or fallback to os.getcwd()
working_dir_str = nix_vars.get("WORKING_DIR", os.path.abspath(os.getcwd()))
working_dir = Path(working_dir_str)

# Use CONF_DIR from Nix vars or fallback to default relative path
conf_dir_rel = nix_vars.get("CONF_DIR", "conf")
conf_dir = (working_dir / conf_dir_rel).resolve()
db_pwd_file = conf_dir / "db_pwd"

# Update nix_vars with resolved absolute paths for consistency if needed elsewhere
nix_vars["WORKING_DIR"] = str(working_dir)
nix_vars["CONF_DIR"] = str(conf_dir) # Store absolute conf path
home_dir = nix_vars.get("HOME_DIR", os.path.expanduser("~")) # Keep home_dir logic
nix_vars["HOME_DIR"] = home_dir

# --- Generate Secrets ---
SALT = get_random_secret_key()
SECRET_KEY = get_random_secret_key()

# --- Ensure conf dir and db_pwd file exist ---
print(f"Checking configuration directory: {conf_dir}")
if not conf_dir.exists():
    print(f"Creating configuration directory: {conf_dir}")
    conf_dir.mkdir(parents=True, exist_ok=True)
else:
    print("Configuration directory already exists.")

print(f"Checking database password file: {db_pwd_file}")
if not db_pwd_file.exists():
    print(f"Database password file not found. Creating '{db_pwd_file}' with default password.")
    try:
        with open(db_pwd_file, 'w', encoding='utf-8') as f:
            f.write(DEFAULT_DB_PASSWORD)
        print(f"Successfully created '{db_pwd_file}'. IMPORTANT: Change the default password for production!")
    except IOError as e:
        print(f"ERROR: Failed to create database password file '{db_pwd_file}': {e}")
else:
    print("Database password file already exists.")


# --- Manage .env file ---
template = Path("./conf/default.env")
target = Path(".env") # .env should be in the working_dir (project root)

# Create a new .env file from template if it doesn't exist
if not target.exists():
    print(f"Creating .env file from template: {template}")
    try:
        shutil.copy(template, target)
    except Exception as e:
        print(f"Error copying template {template} to {target}: {e}")
else:
    print(".env file already exists. Updating...")

# Track what we've found or added in .env
found_keys = set()

# Read existing entries from .env
lines = []
if target.exists():
    try:
        with target.open("r", encoding="utf-8") as f:
            lines = f.readlines()
    except IOError as e:
        print(f"Error reading .env file {target}: {e}")


# Process and update entries
updated_lines = []
django_module_from_nix = nix_vars.get("DJANGO_MODULE")

for line in lines:
    stripped_line = line.strip()
    if not stripped_line or stripped_line.startswith("#"):
        updated_lines.append(line)
        continue

    if "=" not in stripped_line:
        updated_lines.append(line)
        continue

    key, value = stripped_line.split("=", 1)
    key = key.strip()
    found_keys.add(key)

    # Replace specific values from nix_vars if present, without quotes
    if django_module_from_nix:
        if key == "DJANGO_SETTINGS_MODULE":
            updated_lines.append(f'{key}={django_module_from_nix}.settings_dev\n')
            continue
        elif key == "DJANGO_SETTINGS_MODULE_PRODUCTION":
            updated_lines.append(f'{key}={django_module_from_nix}.settings_prod\n')
            continue
        elif key == "DJANGO_SETTINGS_MODULE_DEVELOPMENT":
            updated_lines.append(f'{key}={django_module_from_nix}.settings_dev\n')
            continue

    # Keep existing line if no specific update rule matched
    updated_lines.append(line)


# Write updated content back to .env
try:
    with target.open("w", encoding="utf-8") as f:
        f.writelines(updated_lines)
except IOError as e:
    print(f"Error writing updated .env file {target}: {e}")

# Add any missing required entries to .env without quotes
try:
    with target.open("a", encoding="utf-8") as f:
        # Add secrets if missing
        if "DJANGO_SECRET_KEY" not in found_keys:
            f.write(f'\nDJANGO_SECRET_KEY={SECRET_KEY}') # No quotes
            print("Added DJANGO_SECRET_KEY to .env")

        if "DJANGO_SALT" not in found_keys:
            f.write(f'\nDJANGO_SALT={SALT}') # No quotes
            print("Added DJANGO_SALT to .env")
        
        # Add Storage_DIR if missing
        if "STORAGE_DIR" not in found_keys:
            storage_dir = nix_vars.get("STORAGE_DIR", str(working_dir / "storage"))
            f.write(f'\nSTORAGE_DIR={storage_dir}')
            print("Added STORAGE_DIR to .env")

        # Add paths and config from nix_vars if missing
        # Ensure paths are NOT quoted
        #TODO Check for missing paths from path utils
        vars_to_add = {
            "DJANGO_HOST": nix_vars.get("HOST"),
            "DJANGO_PORT": nix_vars.get("PORT"),
            "DJANGO_CONF_DIR": str(conf_dir),
            "HOME_DIR": nix_vars.get("HOME_DIR"),
            "WORKING_DIR": nix_vars.get("WORKING_DIR"),
            "DJANGO_DATA_DIR": str(working_dir / nix_vars.get("DATA_DIR", "data")),
            "DJANGO_IMPORT_DATA_DIR": str(working_dir / nix_vars.get("IMPORT_DIR", "data/import")),
            "DJANGO_VIDEO_IMPORT_DATA_DIR": str(working_dir / nix_vars.get("IMPORT_DIR", "data/import") / "video"),
            "VIDEO_ALLOW_FPS_FALLBACK": "True",
            "VIDEO_DEFAULT_FPS": "50",
        }
        for key, value in vars_to_add.items():
            if value is not None and key not in found_keys:
                f.write(f'\n{key}={value}') # No quotes
                print(f"Added {key} to .env")

        # Add Django settings module variants if missing and module name is known
        if django_module_from_nix:
            settings_variants = {
                "DJANGO_SETTINGS_MODULE": f"{django_module_from_nix}.settings_dev",
                "DJANGO_SETTINGS_MODULE_PRODUCTION": f"{django_module_from_nix}.settings_prod",
                "DJANGO_SETTINGS_MODULE_DEVELOPMENT": f"{django_module_from_nix}.settings_dev",
            }
            for key, value in settings_variants.items():
                if key not in found_keys:
                    f.write(f'\n{key}={value}') # No quotes
                    print(f"Added {key} to .env")

        # Add other defaults if missing
        default_values = {
            "TEST_RUN": "False",
            "TEST_RUN_FRAME_NUMBER": "1000",
            "RUST_BACKTRACE": "1",
            "DJANGO_DEBUG": "True",
            "DJANGO_FFMPEG_EXTRACT_FRAME_BATCHSIZE": "500",
            "LABEL_VIDEO_SEGMENT_MIN_DURATION_S_FOR_ANNOTATION": "3" # Added missing default
        }
        for key, value in default_values.items():
            if key not in found_keys:
                f.write(f'\n{key}={value}') # No quotes
                print(f"Added {key} to .env")

except IOError as e:
    print(f"Error appending missing entries to .env file {target}: {e}")


print(f"Environment setup script finished. Check {target} and {db_pwd_file}")



from ..utils.translation import build_multilingual_response
from .anonymization import (
    AnonymizationOverviewView,
    AnonymizationValidateView,
    anonymization_current,
    anonymization_status,
    start_anonymization,
)
from .auth import KeycloakVideoView, keycloak_callback, keycloak_login, public_home
from .examination import (
    ExaminationManifestCache,
    ExaminationViewSet,
    get_classification_choices_for_examination,
    get_classifications_for_examination,
    get_findings_for_examination,
    get_instruments_for_examination,
    get_interventions_for_examination,
    get_location_classification_choices_for_examination,
    get_location_classifications_for_examination,
    get_morphology_classification_choices_for_examination,
    get_morphology_classifications_for_examination,
)
from .finding import (
    FindingViewSet,
    get_classifications_for_finding,
    get_interventions_for_finding,
)
from .finding_classification import (
    FindingClassificationViewSet,
    get_classification_choices,
    get_location_choices,  # DEPRECATED
    get_morphology_choices,  # DEPRECATED
)
from .media import (
    get_sensitive_metadata_pk,
    label_list,
    pdf_sensitive_metadata,
    pdf_sensitive_metadata_list,
    pdf_sensitive_metadata_verify,
    sensitive_metadata_list,
    video_sensitive_metadata,
    video_sensitive_metadata_verify,
)
from .meta import SensitiveMetaListView, SensitiveMetaVerificationView
from .misc import (
    MODELTRANSLATION_SETTINGS,
    CenterViewSet,
    ExaminationStatsView,
    ExaminationTranslationOptions,
    FindingClassificationChoiceTranslationOptions,
    FindingClassificationTranslationOptions,
    FindingTranslationOptions,
    GenderViewSet,
    GeneralStatsView,
    InterventionTranslationOptions,
    SensitiveMetaStatsView,
    TranslatedFieldMixin,
    TranslatedFixtureLoader,
    TranslationMigrationHelper,
    UploadFileView,
    UploadStatusView,
    VideoSegmentStatsView,
    csrf_token_view,
)
from .patient import PatientViewSet
from .patient_examination import (
    ExaminationCreateView,
    PatientExaminationDetailView,
    PatientExaminationListView,
    PatientExaminationViewSet,
)
from .patient_finding import OptimizedPatientFindingViewSet, PatientFindingViewSet
from .patient_finding_classification import create_patient_finding_classification
from .report import ReportReimportView, ReportStreamView
from .requirement import LookupViewSet, evaluate_requirements
from .video import (  # Video Correction (Phase 1.1) - Implemented; Existing views
    VideoApplyMaskView,
    VideoCorrectionView,
    VideoExaminationViewSet,
    VideoProcessingHistoryView,
    VideoReimportView,
    VideoRemoveFramesView,
    VideoStreamView,
)

__all__ = [
    # Anonymization views
    "anonymization_status",
    "anonymization_current",
    "start_anonymization",
    "AnonymizationOverviewView",
    "AnonymizationValidateView",
    # Auth views
    "KeycloakVideoView",
    "keycloak_login",
    "keycloak_callback",
    "public_home",
    # Examination views
    "ExaminationManifestCache",
    "ExaminationViewSet",
    "get_classification_choices_for_examination",
    "get_morphology_classification_choices_for_examination",
    "get_location_classification_choices_for_examination",
    "get_classifications_for_examination",
    "get_location_classifications_for_examination",
    "get_morphology_classifications_for_examination",
    "get_findings_for_examination",
    "get_instruments_for_examination",
    "get_interventions_for_examination",
    # Finding Views
    "FindingViewSet",
    "get_interventions_for_finding",
    "get_classifications_for_finding",
    # Finding Classification Views
    "FindingClassificationViewSet",
    "get_classification_choices",
    "get_morphology_choices",  # DEPRECATED
    "get_location_choices",  # DEPRECATED
    # Meta Views
    "SensitiveMetaListView",
    "SensitiveMetaVerificationView",
    # Misc
    "CenterViewSet",
    "csrf_token_view",
    "GenderViewSet",
    "ExaminationStatsView",
    "VideoSegmentStatsView",
    "SensitiveMetaStatsView",
    "GeneralStatsView",
    "ExaminationTranslationOptions",
    "FindingTranslationOptions",
    "FindingClassificationTranslationOptions",
    "FindingClassificationChoiceTranslationOptions",
    "InterventionTranslationOptions",
    "TranslatedFieldMixin",
    "TranslationMigrationHelper",
    "TranslatedFixtureLoader",
    "build_multilingual_response",
    "MODELTRANSLATION_SETTINGS",
    "UploadFileView",
    "UploadStatusView",
    # Patient Views
    "PatientViewSet",
    # Patient Examination Views
    "ExaminationCreateView",
    "PatientExaminationDetailView",
    "PatientExaminationListView",
    "PatientExaminationViewSet",
    # Patient Finding Views
    "PatientFindingViewSet",
    "OptimizedPatientFindingViewSet",
    # Patient Finding Classification Views
    "create_patient_finding_classification",
    # report
    "ReportReimportView",
    "ReportStreamView",
    # Requirement Views
    "evaluate_requirements",
    "LookupViewSet",
    # Video Views (Phase 1.1 - Implemented)
    "VideoProcessingHistoryView",
    "VideoApplyMaskView",
    "VideoRemoveFramesView",
    "VideoCorrectionView",
    "VideoReimportView",
    "VideoStreamView",
    "VideoExaminationViewSet",
    "ReportReimportView",
    "label_list",
    "get_sensitive_metadata_pk",
    "video_sensitive_metadata",
    "video_sensitive_metadata_verify",
    "pdf_sensitive_metadata",
    "pdf_sensitive_metadata_verify",
    "sensitive_metadata_list",
    "pdf_sensitive_metadata_list",
]



from endoreg_db.models import Patient


from rest_framework import serializers


class PatientDropdownSerializer(serializers.ModelSerializer):
    """Serializer für Patient-Dropdown"""
    display_name = serializers.SerializerMethodField()

    class Meta:
        model = Patient
        fields = ['id', 'patient_hash', 'first_name', 'last_name', 'display_name', 'dob']

    def get_display_name(self, obj):
        """
        Returns a user-friendly display string for a patient, combining their first and last name with a shortened patient hash or a placeholder if the hash is missing.
        
        Parameters:
            obj: The patient instance being serialized.
        
        Returns:
            str: The formatted display name for the patient.
        """
        patient_hash = obj.patient_hash
        hash_display = f"({patient_hash[:8]}...)" if patient_hash else "(No Hash)"
        return f"{obj.first_name} {obj.last_name} {hash_display}"



import os

import yaml
from django.conf import settings
from django.core.management.base import BaseCommand

from ...data import REPORT_TYPE_DATA_DIR

#### CUSTOMIZE
from ...models import PdfType, ReportReaderFlag
from ...utils import load_model_data_from_yaml

SOURCE_DIR = REPORT_TYPE_DATA_DIR  # e.g. settings.DATA_DIR_INTERVENTION

MODEL_0 = PdfType

IMPORT_MODELS = [  # string as model key, serves as key in IMPORT_METADATA
    MODEL_0.__name__,
]

IMPORT_METADATA = {
    MODEL_0.__name__: {
        "dir": SOURCE_DIR,  # e.g. "interventions"
        "model": MODEL_0,  # e.g. Intervention
        "dir": SOURCE_DIR,  # e.g. "interventions"
        "model": MODEL_0,
        "foreign_keys": [
            "patient_info_line",
            "endoscope_info_line",
            "examiner_info_line",
            "cut_off_below_lines",
            "cut_off_above_lines"
        ], # e.g. ["intervention_types"]
        "foreign_key_models": [
            ReportReaderFlag,
            ReportReaderFlag,
            ReportReaderFlag,
            ReportReaderFlag,
            ReportReaderFlag
        ] # e.g. [InterventionType]
    }
}


class Command(BaseCommand):
    help = f"""Load all .yaml files in the {SOURCE_DIR} directory"""

    def add_arguments(self, parser):
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="Display verbose output",
        )

    def handle(self, *args, **options):
        verbose = options["verbose"]
        verbose = options["verbose"]
        for model_name in IMPORT_MODELS:
            _metadata = IMPORT_METADATA[model_name]
            load_model_data_from_yaml(self, model_name, _metadata, verbose)

            load_model_data_from_yaml(self, model_name, _metadata, verbose)



import numbers
from typing import TYPE_CHECKING, Optional

from django.db import models

if TYPE_CHECKING:
    from endoreg_db.utils.links.requirement_link import RequirementLinks  # Added import

    from ...administration.person.patient import Patient
    from ...other.unit import Unit
    from ..laboratory import LabValue
    from .patient_lab_sample import PatientLabSample


class PatientLabValue(models.Model):
    """
    A class representing a patient lab value.

    Attributes:
        patient (Patient): The patient.
        lab_value (LabValue): The lab value.
        value (float): The value of the lab value.
        date (datetime): The date of the lab value.
    """

    patient = models.ForeignKey("Patient", on_delete=models.CASCADE, related_name="lab_values", blank=True, null=True)
    lab_value = models.ForeignKey("LabValue", on_delete=models.CASCADE)
    value = models.FloatField(blank=True, null=True)
    value_str = models.CharField(max_length=255, blank=True, null=True)
    sample = models.ForeignKey("PatientLabSample", on_delete=models.CASCADE, blank=True, null=True, related_name="values")
    datetime = models.DateTimeField(  # if not set, use now
        auto_now_add=True
    )
    normal_range = models.JSONField(default=dict)
    unit = models.ForeignKey("Unit", on_delete=models.CASCADE, blank=True, null=True)

    @property
    def lab_value_safe(self) -> "LabValue":
        """Returns the lab value, raises error if not set."""
        if not self.lab_value:
            raise ValueError("Lab value is not set.")
        return self.lab_value

    @property
    def patient_safe(self) -> "Patient":
        """Returns the patient, raises error if not set."""
        if not self.patient:
            raise ValueError("Patient is not set.")
        return self.patient

    if TYPE_CHECKING:
        patient: models.ForeignKey["Patient|None"]
        lab_value: models.ForeignKey["LabValue"]
        unit: models.ForeignKey["Unit|None"]
        sample: models.ForeignKey["PatientLabSample|None"]
        normal_range: models.JSONField[dict]

    @classmethod
    def create_lab_value_by_sample(
        cls, sample: "PatientLabSample", lab_value_name: str, value: Optional[float] = None, value_str: Optional[str] = None, unit: Optional["Unit"] = None
    ):
        from ..laboratory import LabValue

        patient = sample.patient
        lab_value = LabValue.objects.get(name=lab_value_name)

        pat_lab_val = cls.objects.create(
            patient=patient,
            lab_value=lab_value,
            value=value,
            value_str=value_str,
            sample=sample,
            unit=unit,
        )

        pat_lab_val.save()

        return pat_lab_val

    def __str__(self):
        formatted_datetime = self.datetime.strftime("%Y-%m-%d %H:%M")
        # normal_range = self.get_normal_range()
        norm_range_string = f"[{self.normal_range.get('min', '')} - {self.normal_range.get('max', '')}]"
        _str = f"{self.lab_value} - {self.value} {self.unit} - {norm_range_string} ({formatted_datetime})"
        return _str

    def get_normal_range(self):
        lab_value = self.lab_value_safe
        patient = self.patient_safe

        age = patient.age_safe
        gender = patient.gender

        normal_range_dict = lab_value.get_normal_range(age, gender)
        return normal_range_dict

    def set_min_norm_value(self, value, save=True):
        self.normal_range["min"] = value
        if save:
            self.save()

    def set_max_norm_value(self, value, save=True):
        self.normal_range["max"] = value
        if save:
            self.save()

    def set_norm_values_from_default(self):
        normal_range_dict = self.get_normal_range()
        self.set_min_norm_value(normal_range_dict["min"], save=False)
        self.set_max_norm_value(normal_range_dict["max"], save=False)
        self.save()

    def set_unit_from_default(self):
        _default_unit = self.lab_value_safe.default_unit
        self.unit = _default_unit
        self.save()

    def get_value(self):
        if self.value:
            return self.value
        else:
            return self.value_str

    def get_value_field_name(self):
        if self.value:
            return "value"
        else:
            return "value_str"

    # customize save method so that if a numeric value exists, we round it to the precision of the lab value
    def save(self, *args, **kwargs):
        if self.value is not None:
            # only attempt rounding for real numeric types (ints/floats/compatible)

            precision = getattr(self.lab_value_safe, "numeric_precision", None)
            if isinstance(self.value, numbers.Real) and precision is not None:
                # ensure a plain float is passed to built-in round to satisfy type checkers
                self.value = round(float(self.value), int(precision))
        super().save(*args, **kwargs)

    def set_value_by_distribution(self, distribution=None, save=True):
        import warnings

        from ...other.distribution import (
            DateValueDistribution,
            MultipleCategoricalValueDistribution,
            NumericValueDistribution,
            SingleCategoricalValueDistribution,
        )

        patient = self.patient_safe
        lab_value = self.lab_value_safe

        self.unit = self.lab_value_safe.default_unit

        if not distribution:
            distribution = lab_value.get_default_default_distribution()

            if not distribution:
                warnings.warn(f"No distribution set for lab value {lab_value}, assuming uniform numeric distribution based on normal values")

                if not self.normal_range.get("min", None) or not self.normal_range.get("max", None):
                    self.set_norm_values_from_default()
                _min = self.normal_range.get("min", 0.0001)
                _max = self.normal_range.get("max", 100)
                _name = "auto-" + self.lab_value_safe.name + "-distribution-default-uniform"
                distribution = NumericValueDistribution(name=_name, min_descriptor=_min, max_max_desciptor=_max, distribution_type="uniform")

                value = distribution.generate_value(lab_value=lab_value, patient=patient)
                self.value = value
                if save:
                    self.save()

                return value

        if isinstance(distribution, SingleCategoricalValueDistribution):
            value = distribution.generate_value()
            self.value_str = value
            if save:
                self.save()
            return value

        elif isinstance(distribution, NumericValueDistribution):
            value = distribution.generate_value(lab_value=lab_value, patient=patient)
            self.value = value
            if save:
                self.save()
            return value

        elif isinstance(distribution, MultipleCategoricalValueDistribution):
            value = distribution.generate_value()
            self.value_str = value
            if save:
                self.save()
            return value

        elif isinstance(distribution, DateValueDistribution):
            # raise not implemented error
            value = distribution.generate_value()
            self.value = value
            if save:
                self.save()
            return value

    @property
    def links(self) -> "RequirementLinks":
        """
        Aggregates and returns all related model instances relevant for requirement evaluation
        as a RequirementLinks object.
        """
        from endoreg_db.utils.links.requirement_link import RequirementLinks

        return RequirementLinks(
            patient_lab_values=[self]  # Include the lab value itself
        )



from .examination import Examination, ExaminationRequirementSet
from .examination_indication import ExaminationIndication, ExaminationIndicationClassification, ExaminationIndicationClassificationChoice
from .examination_time import ExaminationTime
from .examination_time_type import ExaminationTimeType
from .examination_type import ExaminationType

__all__ = [
    "Examination",
    "ExaminationRequirementSet",
    "ExaminationType",
    "ExaminationTime",
    "ExaminationTimeType",
    "ExaminationIndication",
    "ExaminationIndicationClassification",
    "ExaminationIndicationClassificationChoice",
]



from typing import TYPE_CHECKING

from django.db import models

from .abstract import AbstractState


class LabelVideoSegmentState(AbstractState):
    """State for label video segment data."""

    prediction = models.BooleanField(default=False)
    annotation = models.BooleanField(default=False)
    frames_extracted = models.BooleanField(default=False)
    is_validated = models.BooleanField(default=False)

    origin = models.OneToOneField(
        "LabelVideoSegment",
        on_delete=models.CASCADE,
        related_name="state",
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name = "Label Video Segment State"
        verbose_name_plural = "Label Video Segment States"

    if TYPE_CHECKING:
        from endoreg_db.models import LabelVideoSegment

        origin: models.OneToOneField["LabelVideoSegment|None"]



# processing/__init__.py

from . import sensitive_meta_adapter
from . import report_processing
from . import video_processing

__all__ = [
    "sensitive_meta_adapter",
    "report_processing",
    "video_processing",
]



from PIL import Image
import numpy as np
from typing import List, Tuple, Union, Dict, Callable, Optional
import pytesseract
from .custom_logger import logger
from .ocr_preprocessing import preprocess_image


def pyramid_ocr(image: Image.Image,
               scales: Optional[List[float]] = None,
               preprocessing_methods: Optional[List[str]] = None) -> Tuple[str, Dict]:
    """
    Process an image at multiple scales and combine results for best OCR output.
    
    Parameters:
        image: PIL.Image - Input image
        scales: List[float] - List of scaling factors (e.g., [0.5, 1.0, 1.5, 2.0])
        preprocessing_methods: List[str] - List of preprocessing methods to apply
        
    Returns:
        str - Combined OCR text
        Dict - Dictionary with metadata about OCR process
    """
    if scales is None:
        scales = [0.8, 1.0, 1.5, 2.0]
    
    if preprocessing_methods is None:
        preprocessing_methods = ['grayscale', 'contrast', 'sharpen']
    
    results = []
    confidences = []
    metadata: Dict[str, Any] = {"scale_results": {}}

    for scale in scales:
        logger.info(f"Processing image at scale {scale}")
        width, height = image.size
        new_size = (int(width * scale), int(height * scale))
        
        # Resize image to current scale
        resized_image = image.resize(new_size, Image.Resampling.LANCZOS)

        # Apply preprocessing
        processed_image = preprocess_image(resized_image, preprocessing_methods)
        
        # Apply Tesseract OCR with optimized parameters
        try:
            # Use a page segmentation mode appropriate for the scale
            # PSM 6: Assume a single uniform block of text
            # PSM 3: Fully automatic page segmentation, but no OSD
            psm = 6 if scale <= 1.0 else 3
            config = f'--oem 1 --psm {psm}'
            
            # Extract text and data for confidence calculation
            text = pytesseract.image_to_string(processed_image, config=config)
            data = pytesseract.image_to_data(processed_image, output_type=pytesseract.Output.DICT, config=config)
            
            # Calculate average confidence, excluding negative values
            confidences_list = [float(conf) for conf in data['conf'] if conf != '-1' and conf.strip()]
            avg_confidence = sum(confidences_list) / len(confidences_list) if confidences_list else 0
            
            # Store result for this scale
            results.append(text)
            confidences.append(avg_confidence)
            
            # Add to metadata
            metadata["scale_results"][str(scale)] = {
                "confidence": avg_confidence,
                "text_length": len(text),
                "word_count": len(text.split())
            }
            
            logger.debug(f"Scale {scale} - Confidence: {avg_confidence:.2f}, Text length: {len(text)}")
            
        except Exception as e:
            logger.error(f"OCR processing failed at scale {scale}: {str(e)}")
            results.append("")
            confidences.append(0)
            metadata["scale_results"][str(scale)] = {"error": str(e)}
    
    # Choose best scale based on confidence
    best_index = confidences.index(max(confidences)) if any(confidences) else 0
    best_scale = scales[best_index]
    best_text = results[best_index]
    
    logger.info(f"Best scale: {best_scale} with confidence {confidences[best_index]:.2f}")
    
    # Add summary to metadata
    metadata["selected_scale"] = best_scale
    metadata["selected_confidence"] = confidences[best_index]
    metadata["all_confidences"] = dict(zip(map(str, scales), confidences))
    
    return best_text, metadata


def segment_and_ocr(image: Image.Image,
                   segmentation_mode: str = 'blocks',
                   preprocessing_methods: Optional[List[str]] = None) -> str:
    """
    Segment the image into logical blocks and apply OCR to each segment.
    
    Different segmentation modes:
    - 'blocks': Use Tesseract to find text blocks
    - 'table': Detect and handle tables differently
    - 'layout': Use more advanced layout analysis
    
    Parameters:
        image: PIL.Image - Input image
        segmentation_mode: str - How to segment the image
        preprocessing_methods: List[str] - List of preprocessing methods
        
    Returns:
        str - Combined OCR text from all segments
    """
    if preprocessing_methods is None:
        preprocessing_methods = ['grayscale', 'contrast']
    
    # Apply basic preprocessing
    processed_image = preprocess_image(image, preprocessing_methods)
    
    # Get image dimensions
    width, height = processed_image.size
    
    if segmentation_mode == 'blocks':
        # Use Tesseract to find text blocks
        data = pytesseract.image_to_data(processed_image, output_type=pytesseract.Output.DICT)

        # Group words into blocks based on block_num
        blocks: Dict[int, List[Dict[str, Any]]] = {}
        for i in range(len(data['text'])):
            if data['text'][i].strip():
                block_num = data['block_num'][i]
                if block_num not in blocks:
                    blocks[block_num] = []
                blocks[block_num].append({
                    'text': data['text'][i],
                    'left': data['left'][i],
                    'top': data['top'][i],
                    'width': data['width'][i],
                    'height': data['height'][i]
                })
        
        # Sort blocks by vertical position (top to bottom)
        sorted_blocks = sorted(blocks.values(), key=lambda block: min(word['top'] for word in block))
        
        # Extract text from each block
        text_blocks = []
        for block in sorted_blocks:
            block_text = ' '.join(word['text'] for word in block)
            text_blocks.append(block_text)
        
        return '\n'.join(text_blocks)
    
    elif segmentation_mode == 'table':
        # Convert to numpy array for OpenCV operations
        import cv2
        img_np = np.array(processed_image)
        
        # Convert to grayscale if needed
        if len(img_np.shape) == 3:
            img_np = cv2.cvtColor(img_np, cv2.COLOR_RGB2GRAY)
        
        # Detect lines for table structure
        horizontal_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (40, 1))
        vertical_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1, 40))
        
        # Detect horizontal lines
        horizontal_lines = cv2.morphologyEx(img_np, cv2.MORPH_OPEN, horizontal_kernel)
        
        # Detect vertical lines
        vertical_lines = cv2.morphologyEx(img_np, cv2.MORPH_OPEN, vertical_kernel)
        
        # Combine lines to get table structure
        table_structure = cv2.add(horizontal_lines, vertical_lines)
        
        # Find contours of cells
        contours, _ = cv2.findContours(table_structure, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        # Process each cell
        cell_texts = []
        for contour in contours:
            x, y, w, h = cv2.boundingRect(contour)
            cell_image = processed_image.crop((x, y, x+w, y+h))
            cell_text = pytesseract.image_to_string(cell_image)
            cell_texts.append(cell_text.strip())
        
        return '\n'.join(cell_texts)
    
    elif segmentation_mode == 'layout':
        # Simplified layout analysis - divide into rows
        row_height = height // 10  # Divide image into 10 rows
        
        row_texts = []
        for i in range(10):
            y_start = i * row_height
            y_end = (i + 1) * row_height
            
            # Skip empty rows
            if y_end > height:
                continue
                
            row_image = processed_image.crop((0, y_start, width, y_end))
            row_text = pytesseract.image_to_string(row_image)
            
            if row_text.strip():
                row_texts.append(row_text.strip())
        
        return '\n'.join(row_texts)
    
    else:
        # Fallback - process entire image
        return pytesseract.image_to_string(processed_image)


def adaptive_multi_engine_ocr(image: Image.Image) -> str:
    """
    Adaptively choose and apply the best OCR engine based on image characteristics.
    
    Parameters:
        image: PIL.Image - Input image
        
    Returns:
        str - OCR result from the best engine
    """
    # Analyze image to determine characteristics
    width, height = image.size
    is_high_res = width > 2000 or height > 2000
    
    # Convert to grayscale for analysis
    gray_img = image.convert('L')
    img_array = np.array(gray_img)
    
    # Check if image has low contrast
    min_val, max_val = np.min(img_array), np.max(img_array)
    contrast_ratio = (max_val - min_val) / 255
    is_low_contrast = contrast_ratio < 0.3
    
    # Check noise level (standard deviation in small patches)
    noise_level = np.std(img_array)
    is_noisy = noise_level > 20
    
    logger.info(f"Image characteristics - High res: {is_high_res}, "
               f"Low contrast: {is_low_contrast}, Noisy: {is_noisy}")
    
    # Choose best OCR approach based on characteristics
    if is_high_res:
        logger.info("High resolution image detected, using pyramid OCR")
        text, _ = pyramid_ocr(image, scales=[0.5, 0.75, 1.0])
        return text
    
    if is_low_contrast:
        logger.info("Low contrast image detected, applying contrast enhancement")
        from .ocr_preprocessing import preprocess_image
        image = preprocess_image(image, methods=['grayscale', 'contrast', 'threshold'])
    
    if is_noisy:
        logger.info("Noisy image detected, applying denoising")
        from .ocr_preprocessing import preprocess_image
        image = preprocess_image(image, methods=['grayscale', 'denoise'])
    
    # Default: try both Tesseract and TrOCR and pick the best result
    try:
        tesseract_text = pytesseract.image_to_string(image)
    except Exception as e:
        logger.error(f"Tesseract OCR failed: {str(e)}")
        tesseract_text = ""
    
    try:
        from .ocr import trocr_full_image_ocr
        trocr_text = trocr_full_image_ocr(image)
    except Exception as e:
        logger.error(f"TrOCR failed: {str(e)}")
        trocr_text = ""
    
    # Choose result with more content (basic heuristic)
    if len(trocr_text.strip()) > len(tesseract_text.strip()):
        logger.info("TrOCR produced better results (more content)")
        return trocr_text
    else:
        logger.info("Tesseract produced better results (more content)")
        return tesseract_text



# services/lookup_store.py

from __future__ import annotations

import uuid
from typing import Any, Dict, Iterable, Optional

from django.conf import settings
from django.core.cache import cache

from endoreg_db.models.medical import PatientExamination

# Align TTL with Django cache TIMEOUT for consistency in tests and runtime


try:
    DEFAULT_TTL_SECONDS = int(
        settings.CACHES.get("default", {}).get("TIMEOUT", 60 * 30)
    )


except Exception:
    DEFAULT_TTL_SECONDS = 60 * 30  # 30 minutes fallback


class LookupStore:
    """

    Server-side lookup dictionary stored in Django cache.

    Return a token to the client; later requests use that token to get/update parts.

    """

    def __init__(self, token: Optional[str] = None):
        self.token = token or uuid.uuid4().hex

    @property
    def cache_key(self) -> str:
        return f"lookup:{self.token}"

    def init(
        self, initial: Optional[Dict[str, Any]] = None, ttl: int = DEFAULT_TTL_SECONDS
    ) -> str:
        cache.set(self.cache_key, initial or {}, ttl)

        return self.token

    def get_all(self) -> Dict[str, Any]:
        return cache.get(self.cache_key, {})

    def get_many(self, keys: Iterable[str]) -> Dict[str, Any]:
        data = self.get_all()

        return {k: data.get(k) for k in keys}

    def set_many(self, updates: Dict[str, Any], ttl: int = DEFAULT_TTL_SECONDS) -> None:
        data = self.get_all()

        data.update(updates)

        cache.set(self.cache_key, data, ttl)

    def get(self, key: str, default: Any = None) -> Any:
        return self.get_all().get(key, default)

    def set(self, key: str, value: Any, ttl: int = DEFAULT_TTL_SECONDS) -> None:
        data = self.get_all()

        data[key] = value

        cache.set(self.cache_key, data, ttl)

    def delete(self) -> None:
        cache.delete(self.cache_key)

    def patch(self, updates: Dict[str, Any], ttl: int = DEFAULT_TTL_SECONDS) -> None:
        """Patch existing data with updates."""

        data = self.get_all()

        data.update(updates)

        cache.set(self.cache_key, data, ttl)

    def validate_and_recover_data(self, token):
        """Validate stored data and attempt recovery if corrupted"""

        data = self.get_all()

        if not data:
            return None

        # Check if required fields are present

        required_fields = [
            "patient_examination_id",
            "requirements_by_set",
            "requirement_status",
        ]

        missing_fields = [field for field in required_fields if field not in data]

        if missing_fields:
            import logging

            logger = logging.getLogger(__name__)

            logger.warning(
                f"Missing fields in lookup data for token {token}: {missing_fields}"
            )

            # Try to recover patient_examination_id from token or related data

            if "patient_examination_id" in missing_fields:
                # Attempt to extract from token or find related examination

                recovered_id = self._recover_patient_examination_id(token)

                if recovered_id:
                    data["patient_examination_id"] = recovered_id

                    logger.info(
                        f"Recovered patient_examination_id {recovered_id} for token {token}"
                    )

            # Do not automatically recompute here to avoid loops

            # Recompute is only triggered by PATCH or explicit POST /recompute/

            # For now, just return the data as is

        return data

    def _recover_patient_examination_id(self, token: str) -> Optional[str]:
        """

        Placeholder for recovering patient_examination_id.

        This should be implemented with logic to find the ID based on the token or other data.

        """

        # In a real implementation, you might query a database or another service.

        # For now, we return None as recovery logic is not defined.

        return None

    def should_recompute(self, token):
        """Check if recomputation is needed based on data freshness"""

        data = self.get_all()

        if not data:
            return True

        # Check if we have a last_recompute timestamp

        last_recompute = data.get("_last_recompute")

        if not last_recompute:
            return True

        # Only recompute if it's been more than 30 seconds since last recompute

        # This prevents excessive recomputation while allowing for updates

        from datetime import datetime, timedelta

        try:
            last_recompute_time = datetime.fromisoformat(last_recompute)

            return datetime.now() - last_recompute_time > timedelta(seconds=30)

        except (ValueError, TypeError):
            return True

    def mark_recompute_done(self):
        """Mark that recomputation has been completed"""

        from datetime import datetime

        self.set("_last_recompute", datetime.now().isoformat())



from django.core.management.base import BaseCommand

from ...data import REPORT_READER_FLAG_DATA_DIR
from ...models import Shift, ShiftType
from ...utils import load_model_data_from_yaml

SOURCE_DIR = REPORT_READER_FLAG_DATA_DIR  # e.g. settings.DATA_DIR_INTERVENTION

model_0 = ShiftType
model_1 = Shift

IMPORT_MODELS = [  # string as model key, serves as key in IMPORT_METADATA
    model_0.__name__,
]

IMPORT_METADATA = {
    model_0.__name__: {
        "dir": SOURCE_DIR,  # e.g. "interventions"
        "model": model_0,
        "foreign_keys": [],  # e.g. ["intervention_types"]
        "foreign_key_models": [],  # e.g. [InterventionType]
    },
    model_1.__name__: {
        "dir": SOURCE_DIR,  # e.g. "interventions"
        "model": model_1,
        "foreign_keys": ["shift_types"],  # e.g. ["intervention_types"]
        "foreign_key_models": [model_0],  # e.g. [InterventionType]
    },
}


class Command(BaseCommand):
    help = """Load all .yaml files in the data/shift directory
    into the Shift and ShiftType model"""

    def add_arguments(self, parser):
        """
        Adds the --verbose command-line option to enable detailed output during execution.
        """
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="Display verbose output",
        )

    def handle(self, *args, **options):
        """
        Loads YAML data files into models defined in IMPORT_MODELS using provided metadata.

        Iterates over each model specified in IMPORT_MODELS, retrieves its import metadata,
        and calls the data loading utility to populate the database from YAML files. Supports
        optional verbose output if enabled via command-line options.
        """
        verbose = options["verbose"]
        for model_name in IMPORT_MODELS:
            _metadata = IMPORT_METADATA[model_name]
            load_model_data_from_yaml(self, model_name, _metadata, verbose)



"""
High-performance Frame-specific OCR module using Tesserocr for video processing.

This module provides significantly faster OCR functionality using Tesserocr:
- Pre-loaded Tesseract model for memoization (10-50x faster)
- Direct C++ API access instead of CLI wrapper
- Optimized for video frame batch processing
- Memory-efficient frame handling
- Medical text recognition patterns
- Optimized for endoscopy video overlays

Performance Benefits over pytesseract:
- Model is loaded once at initialization
- No subprocess overhead per frame
- Direct memory access to Tesseract engine
- Batch processing optimizations
"""

import logging
import cv2
import numpy as np
import tesserocr
from PIL import Image, ImageEnhance, ImageFilter
from typing import Dict, Any, Tuple, Optional, List
import re
import threading
import time

logger = logging.getLogger(__name__)

class TesseOCRFrameProcessor:
    """
    High-performance OCR processor using Tesserocr for video frames.
    
    Key Features:
    - Pre-loaded Tesseract model for maximum speed
    - Batch processing for video frames
    - Thread-safe operations
    - Automatic memory management
    - German medical text optimization
    """
    
    def __init__(self, language: str = 'deu+eng'):
        """
        Initialize TesseOCR processor with pre-loaded model.
        
        Args:
            language: Tesseract language code (e.g., 'deu+eng' for German+English)
        """
        self.language = language
        self._lock = threading.Lock()
    
        
        # Initialize Tesseract API - this is the key optimization!
        try:
            self.api = tesserocr.PyTessBaseAPI(lang=language, path=tessdata_path)
            logger.info(f"TesseOCR initialized with language: {language}")
            
            # Set optimal OCR parameters for video frames
            self.api.SetPageSegMode(tesserocr.PSM.SINGLE_BLOCK)  # PSM 6 equivalent
            self.api.SetVariable('tessedit_char_whitelist', 
                               '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-üöäÜÖÄß.:/ ')
            
            # Optimize for speed vs accuracy
            self.api.SetVariable('tessedit_enable_dict_correction', '1')
            self.api.SetVariable('tessedit_enable_bigram_correction', '1')
            
        except Exception as e:
            logger.error(f"Failed to initialize TesseOCR: {e}")
            self.api = None
            raise
        
        # Frame-specific OCR configurations
        self.frame_config = {
            'dpi': 300,
            'min_confidence': 30,  # Lower threshold for video frames
            'high_quality_dpi': 400,
            'high_quality_min_confidence': 50
        }
        
        # Medical text patterns for enhanced detection
        self.medical_patterns = [
            r'Patient[:\s]*([A-Za-zäöüÄÖÜß\s\-]+)',
            r'Pat[\.:\s]*([A-Za-zäöüÄÖÜß\s\-]+)',
            r'geb[\.:\s]*(\d{1,2}\.\d{1,2}\.\d{2,4})',
            r'Fall[nr]*[\.:\s]*(\d+)',
            r'Datum[:\s]*(\d{1,2}\.\d{1,2}\.\d{2,4})',
            r'Geschlecht[:\s]*(männlich|weiblich|m|w|male|female)',
            r'Geburtsdatum[:\s]*(\d{1,2}\.\d{1,2}\.\d{2,4})',
            r'(\d{1,2}\.\d{1,2}\.\d{2,4})',  # Generic date pattern
            r'(\d{2,})',  # Case numbers
        ]
        
        # Performance metrics
        self.processed_frames = 0
        self.total_processing_time = 0.0
        
    def __del__(self):
        """Clean up Tesseract API on destruction."""
        if hasattr(self, 'api') and self.api:
            try:
                self.api.End()
            except Exception as e:
                logger.debug(f"Error cleaning up Tesseract API: {e}")
    
    def preprocess_frame_for_ocr(self, frame: np.ndarray, roi: Optional[Dict[str, Any]] = None) -> Image.Image:
        """
        Preprocess a video frame for optimal OCR results.
        
        Args:
            frame: Input frame as numpy array (BGR format from cv2)
            roi: Optional region of interest to focus OCR on
            
        Returns:
            Preprocessed PIL Image ready for OCR
        """
        try:
            # Convert BGR to RGB for PIL
            if len(frame.shape) == 3:
                frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            else:
                frame_rgb = frame
            
            # Apply ROI cropping if specified
            if roi and self._validate_roi(roi):
                x, y, w, h = roi['x'], roi['y'], roi['width'], roi['height']
                frame_rgb = frame_rgb[y:y+h, x:x+w]
            
            # Convert to PIL Image
            pil_image = Image.fromarray(frame_rgb)
            
            # Resize for better OCR (upscale small text)
            original_size = pil_image.size
            scale_factor = max(2.0, 1200 / max(original_size))
            new_size = (int(original_size[0] * scale_factor), int(original_size[1] * scale_factor))
            pil_image = pil_image.resize(new_size, Image.Resampling.LANCZOS)
            
            # Convert to grayscale
            pil_image = pil_image.convert('L')
            
            # Enhance contrast for text visibility
            enhancer = ImageEnhance.Contrast(pil_image)
            pil_image = enhancer.enhance(2.0)
            
            # Apply sharpening filter
            pil_image = pil_image.filter(ImageFilter.SHARPEN)
            
            # Threshold to clean binary image
            img_array = np.array(pil_image)
            _, binary = cv2.threshold(img_array, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
            
            # Remove noise with morphological operations
            kernel = np.ones((2,2), np.uint8)
            cleaned = cv2.morphologyEx(binary, cv2.MORPH_CLOSE, kernel)
            cleaned = cv2.morphologyEx(cleaned, cv2.MORPH_OPEN, kernel)
            
            return Image.fromarray(cleaned)
            
        except Exception as e:
            logger.error(f"Frame preprocessing failed: {e}")
            # Fallback to simple conversion
            if len(frame.shape) == 3:
                return Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)).convert('L')
            else:
                return Image.fromarray(frame)
    
    def extract_text_from_frame(
        self, 
        frame: np.ndarray, 
        roi: Optional[Dict[str, Any]] = None,
        high_quality: bool = True
    ) -> Tuple[str, float, Dict[str, Any]]:
        """
        Extract text from a video frame using pre-loaded Tesseract model.
        
        Args:
            frame: Input frame as numpy array
            roi: Optional region of interest for OCR
            high_quality: Use high-quality OCR settings
            
        Returns:
            Tuple of (extracted_text, confidence_score, ocr_data)
        """
        if not self.api:
            logger.error("TesseOCR API not initialized")
            return "", 0.0, {}
            
        start_time = time.time()
        
        try:
            with self._lock:  # Thread safety
                # Preprocess frame
                processed_image = self.preprocess_frame_for_ocr(frame, roi)
                
                # Set DPI for better recognition
                dpi = self.frame_config['high_quality_dpi'] if high_quality else self.frame_config['dpi']
                
                # Set image in Tesseract API
                self.api.SetImage(processed_image)
                
                # Extract text - this is much faster than pytesseract!
                extracted_text = self.api.GetUTF8Text().strip()
                
                # Get confidence score
                confidence = self.api.MeanTextConf()
                confidence_normalized = confidence / 100.0 if confidence > 0 else 0.0
                
                # Get detailed word-level information if needed
                ocr_data = {}
                try:
                    # Get bounding boxes and confidences for each word
                    boxes = self.api.GetComponentImages(tesserocr.RIL.WORD, True)
                    words_info = []
                    
                    for (box, text, conf) in boxes:
                        if text.strip() and conf > (self.frame_config['high_quality_min_confidence'] 
                                                   if high_quality else self.frame_config['min_confidence']):
                            words_info.append({
                                'text': text.strip(),
                                'confidence': conf,
                                'bbox': box
                            })
                    
                    ocr_data = {
                        'words': words_info,
                        'processing_time': time.time() - start_time,
                        'dpi': dpi,
                        'method': 'tesserocr'
                    }
                    
                except Exception as e:
                    logger.debug(f"Could not extract detailed OCR data: {e}")
            
            # Update performance metrics
            self.processed_frames += 1
            self.total_processing_time += time.time() - start_time
            
            logger.debug(f"TesseOCR extracted text: '{extracted_text[:50]}...' "
                        f"with confidence {confidence_normalized:.2f} "
                        f"in {time.time() - start_time:.3f}s")
            
            return extracted_text, confidence_normalized, ocr_data
            
        except Exception as e:
            logger.error(f"TesseOCR text extraction failed: {e}")
            return "", 0.0, {}
    
    def extract_text_from_frame_batch(
        self, 
        frames: List[np.ndarray], 
        roi: Optional[Dict[str, Any]] = None,
        high_quality: bool = True
    ) -> List[Tuple[str, float, Dict[str, Any]]]:
        """
        Process multiple frames in batch for maximum efficiency.
        
        Args:
            frames: List of input frames as numpy arrays
            roi: Optional region of interest for OCR
            high_quality: Use high-quality OCR settings
            
        Returns:
            List of tuples (extracted_text, confidence_score, ocr_data)
        """
        if not self.api:
            logger.error("TesseOCR API not initialized")
            return [("", 0.0, {})] * len(frames)
        
        results = []
        batch_start = time.time()
        
        logger.info(f"Processing batch of {len(frames)} frames with TesseOCR...")
        
        for i, frame in enumerate(frames):
            try:
                result = self.extract_text_from_frame(frame, roi, high_quality)
                results.append(result)
                
                if i % 10 == 0 and i > 0:
                    avg_time = (time.time() - batch_start) / (i + 1)
                    logger.debug(f"Processed {i+1}/{len(frames)} frames, "
                               f"avg time: {avg_time:.3f}s/frame")
                    
            except Exception as e:
                logger.error(f"Error processing frame {i}: {e}")
                results.append(("", 0.0, {}))
        
        total_time = time.time() - batch_start
        logger.info(f"Batch processing complete: {len(frames)} frames in {total_time:.2f}s "
                   f"({total_time/len(frames):.3f}s/frame)")
        
        return results
    
    def detect_sensitive_content(
        self, 
        frame: np.ndarray, 
        roi: Optional[Dict[str, Any]] = None
    ) -> Tuple[bool, str, Dict[str, Any]]:
        """
        Detect sensitive medical information in frame using TesseOCR.
        
        Args:
            frame: Input frame as numpy array
            roi: Optional region of interest
            
        Returns:
            Tuple of (has_sensitive_content, extracted_text, pattern_matches)
        """
        try:
            # Extract text using high-quality OCR
            ocr_text, confidence, ocr_data = self.extract_text_from_frame(
                frame, roi, high_quality=True
            )
            
            if not ocr_text or confidence < 0.3:
                return False, "", {}
            
            # Extract medical patterns
            patterns = self.extract_medical_text_patterns(ocr_text)
            
            # Determine if sensitive content is present
            has_sensitive = bool(patterns)
            
            logger.debug(f"Sensitive content detection: {has_sensitive}, "
                        f"patterns found: {len(patterns)}")
            
            return has_sensitive, ocr_text, patterns
            
        except Exception as e:
            logger.error(f"Sensitive content detection failed: {e}")
            return False, "", {}
    
    def extract_medical_text_patterns(self, text: str) -> Dict[str, Any]:
        """
        Extract medical information patterns from text.
        
        Args:
            text: Input text to analyze
            
        Returns:
            Dictionary of extracted medical information
        """
        extracted_info = {
            'patient_name': [],
            'birth_date': [],
            'case_number': [],
            'examination_date': [],
            'gender': [],
            'dates': [],
            'numbers': []
        }
        
        try:
            for pattern in self.medical_patterns:
                matches = re.finditer(pattern, text, re.IGNORECASE)
                for match in matches:
                    matched_text = match.group(1) if match.groups() else match.group(0)
                    
                    # Categorize matches
                    if any(keyword in pattern.lower() for keyword in ['patient', 'pat']):
                        extracted_info['patient_name'].append(matched_text.strip())
                    elif 'geb' in pattern.lower() or 'geburtsdatum' in pattern.lower():
                        extracted_info['birth_date'].append(matched_text.strip())
                    elif 'fall' in pattern.lower():
                        extracted_info['case_number'].append(matched_text.strip())
                    elif 'datum' in pattern.lower():
                        extracted_info['examination_date'].append(matched_text.strip())
                    elif 'geschlecht' in pattern.lower():
                        extracted_info['gender'].append(matched_text.strip())
                    elif re.match(r'\d{1,2}\.\d{1,2}\.\d{2,4}', matched_text):
                        extracted_info['dates'].append(matched_text.strip())
                    elif re.match(r'\d{2,}', matched_text):
                        extracted_info['numbers'].append(matched_text.strip())
            
            # Remove duplicates
            for key in extracted_info:
                extracted_info[key] = list(set(extracted_info[key]))
            
            return extracted_info
            
        except Exception as e:
            logger.error(f"Medical pattern extraction failed: {e}")
            return extracted_info
    
    def _validate_roi(self, roi: Dict[str, Any]) -> bool:
        """Validate ROI parameters."""
        required_keys = ['x', 'y', 'width', 'height']
        return all(key in roi for key in required_keys) and all(
            isinstance(roi[key], (int, float)) and roi[key] >= 0 
            for key in required_keys
        )
    
    def get_performance_stats(self) -> Dict[str, float]:
        """Get performance statistics."""
        if self.processed_frames == 0:
            return {'avg_processing_time': 0.0, 'total_frames': 0}
        
        return {
            'avg_processing_time': self.total_processing_time / self.processed_frames,
            'total_frames': self.processed_frames,
            'total_time': self.total_processing_time,
            'frames_per_second': self.processed_frames / self.total_processing_time if self.total_processing_time > 0 else 0
        }
    
    def reset_performance_stats(self):
        """Reset performance metrics."""
        self.processed_frames = 0
        self.total_processing_time = 0.0


# Global instance for easy access (optional - can be instantiated as needed)
_global_tesseocr_processor = None

def get_global_tesseocr_processor(language: str = 'deu+eng') -> TesseOCRFrameProcessor:
    """
    Get or create global TesseOCR processor instance.
    
    This allows for model reuse across multiple video processing sessions.
    """
    global _global_tesseocr_processor
    
    if _global_tesseocr_processor is None:
        _global_tesseocr_processor = TesseOCRFrameProcessor(language)
    
    return _global_tesseocr_processor


# Convenience function for drop-in replacement
def extract_text_from_frame_fast(
    frame: np.ndarray, 
    roi: Optional[Dict[str, Any]] = None,
    high_quality: bool = True,
    language: str = 'deu+eng'
) -> Tuple[str, float, Dict[str, Any]]:
    """
    Fast text extraction from frame using global TesseOCR processor.
    
    Drop-in replacement for pytesseract-based functions with significant speed improvement.
    """
    processor = get_global_tesseocr_processor(language)
    return processor.extract_text_from_frame(frame, roi, high_quality)



# endoreg_db/serializers/video/video_file_list.py
from typing import Literal, Optional
import logging

from rest_framework import serializers

from endoreg_db.models import VideoFile

logger = logging.getLogger(__name__)


class VideoFileListSerializer(serializers.ModelSerializer):
    """
    Minimal serializer to return only basic video information
    for the video selection dropdown in Vue.js.

    Convention:
        - Serializer methods must NOT raise if video state is missing or invalid.
        - They return safe defaults and log what went wrong.
    """

    # Add computed fields for video status
    status = serializers.SerializerMethodField()
    assignedUser = serializers.SerializerMethodField()
    anonymized = serializers.SerializerMethodField()

    class Meta:
        model = VideoFile
        fields = ["id", "original_file_name", "status", "assignedUser", "anonymized"]

    # --- internal helper -------------------------------------------------
    def _get_video_state(self, obj: VideoFile):
        """
        Best-effort accessor for obj.state.

        Serializer layer must never raise here; it only logs and returns None
        if the state cannot be loaded for any reason.
        """
        try:
            return getattr(obj, "state", None)
        except Exception as exc:  # pragma: no cover - type of error is DB/backend-specific
            logger.warning(
                "VideoFileListSerializer: unable to access state for VideoFile(id=%s): %s",
                getattr(obj, "id", "unknown"),
                exc,
            )
            return None

    # --- public serializer fields ----------------------------------------
    def get_status(
        self, obj: VideoFile
    ) -> Literal["completed", "in_progress", "available"]:
        """
        Determine the processing status of a video file as 'completed',
        'in_progress', or 'available'.

        Contract:
            - Never raises.
            - Missing or invalid state -> treated as 'available'.
        """
        state = self._get_video_state(obj)

        if not state:
            return "available"

        # Use getattr with defaults to tolerate partially populated state objects
        anonymized = getattr(state, "anonymized", False)
        frames_extracted = getattr(state, "frames_extracted", False)

        if anonymized:
            return "completed"
        if frames_extracted:
            return "in_progress"
        return "available"

    def get_assignedUser(self, obj: VideoFile):
        """
        Returns the user assigned to the video, or None if no user is assigned.

        Currently always returns None as user assignment is not implemented.
        """
        # For now return None, can be extended when user assignment is implemented
        return None

    def get_anonymized(self, obj: VideoFile) -> bool:
        """
        Determine whether the video has been anonymized.

        Contract:
            - Never raises.
            - Returns False if state does not exist or cannot be loaded.
        """
        state = self._get_video_state(obj)
        if not state:
            return False

        # getattr to be robust against partially/populated state
        return bool(getattr(state, "anonymized", False))



from django.urls import path
from endoreg_db.views.ai import (label_list, add_label, delete_label, update_label)

url_patterns=[
    # Label interface
    # GET, Returns list of labels
    path(   
        "ai/label-list",
        label_list,
        name="label-list"
        ),
    # POST, updates label by deleting old one and adding new one
    path(
        "ai/label-update",
        update_label,
        name="label-update"
    ),
    # DELETE, deletes a label
    path(
        "ai/label-delete",
        delete_label,
        name="label-delete"
    ),
    # POST, allows adding a new label
    path(
        "ai/label-add",
        add_label,
        name="label-add"
    )  
    

]


"""
Sensitive Region Cropper - Implementierung für das Cropping von Regionen mit sensitivem Text.

Dieses Modul erkennt sensitive Textregionen in PDFs und erstellt gecropte Versionen
der entsprechenden Bereiche für die Anonymisierung.
"""

import re
from typing import List, Tuple, Dict, Optional
from PIL import Image, ImageDraw
from pathlib import Path
from typing import Iterable
from .ocr import tesseract_full_image_ocr
from .pdf_operations import convert_pdf_to_images
from .spacy_extractor import PatientDataExtractor, ExaminerDataExtractor
from .custom_logger import get_logger

logger = get_logger(__name__)


class SensitiveRegionCropper:
    """
    Klasse für das Erkennen und Cropping von sensitiven Textregionen in medizinischen Dokumenten.
    """
    
    def __init__(self, 
                 margin: int = 20,
                 min_region_size: Tuple[int, int] = (100, 30),
                 merge_distance: int = 50):
        """
        Initialisiert den SensitiveRegionCropper.
        
        Args:
            margin: Zusätzlicher Rand um sensitive Bereiche in Pixeln
            min_region_size: Minimale Größe für Crop-Regionen (Breite, Höhe)
            merge_distance: Maximaler Abstand zum Zusammenführen benachbarter Regionen
        """
        self.margin = margin
        self.min_region_size = min_region_size
        self.merge_distance = merge_distance
        
        # Initialisiere Extraktoren für sensitive Daten
        self.patient_extractor = PatientDataExtractor()
        self.examiner_extractor = ExaminerDataExtractor()
        
        self.patient_info = None
        
        # Definiere sensitive Datentypen und ihre Regex-Patterns
        self.sensitive_patterns = {
            'patient_name': r'[A-ZÄÖÜ][a-zäöüß]+\s*,\s*[A-ZÄÖÜ][a-zäöüß]+',
            'birth_date': r'\b\d{1,2}\.\d{1,2}\.\d{4}\b',
            'case_number': r'(?:Fallnummer|Fallnr|Fall\.Nr)[:\s]*(\d+)',
            'social_security': r'\b\d{2}\s?\d{2}\s?\d{2}\s?\d{4}\b',
            'phone_number': r'\b(?:\+49\s?)?(?:\d{3,5}[\s\-]?)?\d{6,8}\b',
            'address': r'[A-ZÄÖÜ][a-zäöüß\s]+(str\.|straße|platz|weg|gasse)\s*\d+',
            'doctor_name': r'(?:Dr\.\s?(?:med\.\s?)?)?[A-ZÄÖÜ][a-zäöüß]+(?:\s+[A-ZÄÖÜ][a-zäöüß]+)?'
        }
    
    # Füge in der Klasse SensitiveRegionCropper hinzu:


    def _enclosing_box(self, boxes: Iterable[Tuple[int, int, int, int]]) -> Tuple[int, int, int, int]:
        """
        Enclosing-Box-Helfer: (x, y, w, h)-Boxen -> (x1, y1, x2, y2).
        """
        boxes = list(boxes)
        if not boxes:
            return (0, 0, 0, 0)
        x1 = min(b[0] for b in boxes)
        y1 = min(b[1] for b in boxes)
        x2 = max(b[0] + b[2] for b in boxes)
        y2 = max(b[1] + b[3] for b in boxes)
        return (x1, y1, x2, y2)

    def _create_bounding_boxes_recursive(
        self,
        boxes: List[Tuple[int, int, int, int]],
        axis: str = "x",
        gap_threshold: Optional[int] = None,
        depth: int = 0,
        max_depth: int = 10,
    ) -> List[Tuple[int, int, int, int]]:
        """
        Rekursive Zerlegung einer Menge von (x, y, w, h)-Boxen in kleinere
        Enclosing-Boxes, indem entlang einer Achse bei ausreichend großen Lücken
        gesplittet wird. Achse wechselt pro Rekursionsschritt (x -> y -> x ...).

        Args:
            boxes: Liste von (x, y, w, h).
            axis: 'x' (horizontal clustern) oder 'y' (vertikal clustern).
            gap_threshold: minimale Lücke (in Pixeln), um zu splitten.
            depth: aktuelle Rekursionstiefe.
            max_depth: Sicherheitslimit gegen Pathologien.

        Returns:
            Liste von (x1, y1, x2, y2)-Bounding-Boxes.
        """
        if not boxes:
            return []
        if len(boxes) == 1 or depth >= max_depth:
            return [self._enclosing_box(boxes)]

        if gap_threshold is None:
            # konservativ: halbiere merge_distance, aber min. 8 px
            gap_threshold = max(8, self.merge_distance // 2)

        # Sortiere Boxen entlang Achse und berechne Lücken
        if axis == "x":
            boxes_sorted = sorted(boxes, key=lambda b: (b[0], b[1]))
            # gap = linker Rand des nächsten - rechter Rand des vorherigen
            gaps = []
            for i in range(len(boxes_sorted) - 1):
                left = boxes_sorted[i]
                right = boxes_sorted[i + 1]
                gap = (right[0]) - (left[0] + left[2])
                gaps.append((gap, i))
        else:  # axis == "y"
            boxes_sorted = sorted(boxes, key=lambda b: (b[1], b[0]))
            # gap = oberer Rand des nächsten - unterer Rand des vorherigen
            gaps = []
            for i in range(len(boxes_sorted) - 1):
                top = boxes_sorted[i]
                bottom = boxes_sorted[i + 1]
                gap = (bottom[1]) - (top[1] + top[3])
                gaps.append((gap, i))

        # Finde größte Lücke über Schwellwert
        split_index = None
        max_gap = -1
        for gap, i in gaps:
            if gap > gap_threshold and gap > max_gap:
                max_gap = gap
                split_index = i

        if split_index is not None:
            # Splitte in zwei Cluster und rekursiv weiter
            left_cluster = boxes_sorted[: split_index + 1]
            right_cluster = boxes_sorted[split_index + 1 :]

            next_axis = "y" if axis == "x" else "x"

            left_boxes = self._create_bounding_boxes_recursive(
                left_cluster, axis=next_axis, gap_threshold=gap_threshold, depth=depth + 1, max_depth=max_depth
            )
            right_boxes = self._create_bounding_boxes_recursive(
                right_cluster, axis=next_axis, gap_threshold=gap_threshold, depth=depth + 1, max_depth=max_depth
            )
            return left_boxes + right_boxes

        # Keine ausreichend große Lücke auf dieser Achse -> Versuche einmal die andere Achse
        if depth == 0:
            other_axis = "y" if axis == "x" else "x"
            alt = self._create_bounding_boxes_recursive(
                boxes, axis=other_axis, gap_threshold=gap_threshold, depth=depth + 1, max_depth=max_depth
            )
            return alt

        # Keine sinnvolle weitere Zerlegung -> eine Enclosing-Box
        return [self._enclosing_box(boxes)]


    def detect_sensitive_regions(self, 
                                image: Image.Image, 
                                word_boxes: List[Tuple[str, Tuple[int, int, int, int]]]) -> List[Tuple[int, int, int, int]]:
        """
        Erkennt sensitive Textregionen basierend auf OCR-Ergebnissen und Regex-Patterns.
        
        Args:
            image: PIL Image Objekt
            word_boxes: Liste von (word, (x, y, width, height)) Tupeln aus OCR
            
        Returns:
            Liste von (x1, y1, x2, y2) Bounding Boxes für sensitive Regionen
        """
        sensitive_regions = []
        full_text = " ".join([word for word, _ in word_boxes])
        
        logger.info(f"Analysiere Text auf sensitive Daten: {full_text[:100]}...")
        
        # 1. Verwende SpaCy-Extraktoren für strukturierte Erkennung
        patient_info = self.patient_extractor(full_text)
        
        # 2. Finde Positionen von sensitiven Daten in den Word-Boxes
        for data_type, pattern in self.sensitive_patterns.items():
            matches = re.finditer(pattern, full_text, re.IGNORECASE)
            
            for match in matches:
                start_pos = match.start()
                end_pos = match.end()
                matched_text = match.group()
                
                logger.debug(f"Gefunden {data_type}: '{matched_text}' an Position {start_pos}-{end_pos}")
                
                # Finde entsprechende Word-Boxes für diesen Text
                region_boxes = self._find_word_boxes_for_text_span(
                    matched_text, word_boxes, start_pos, end_pos
                )
                
                if region_boxes:
                    # Erstelle Bounding Boxes für diese Region
                    bboxes = self._create_bounding_boxes_recursive(region_boxes, axis="x")
                    sensitive_regions.extend(bboxes)  # extend statt append!
        
        # 3. Zusätzliche Erkennung für Patientendaten aus SpaCy-Extraktor
        if patient_info:
            sensitive_regions.extend(
                self._find_patient_data_regions(patient_info, word_boxes, full_text)
            )
        
        # 4. Führe benachbarte Regionen zusammen
        logger.debug(f"Sensitive regions vor Merge: {len(sensitive_regions)} Regionen")
        for i, region in enumerate(sensitive_regions):
            if not isinstance(region, tuple) or len(region) != 4:
                logger.error(f"Invalid region at index {i}: {region} (type: {type(region)})")
        
        # Filtere ungültige Regionen heraus
        valid_regions = [r for r in sensitive_regions if isinstance(r, tuple) and len(r) == 4]
        if len(valid_regions) != len(sensitive_regions):
            logger.warning(f"Filtered out {len(sensitive_regions) - len(valid_regions)} invalid regions")
        
        merged_regions = self._merge_nearby_regions(valid_regions)
        
        # 5. Erweitere Regionen um Margin und validiere Größe
        final_regions = []
        for x1, y1, x2, y2 in merged_regions:
            x1 = max(0, x1 - self.margin)
            y1 = max(0, y1 - self.margin)
            x2 = min(image.width, x2 + self.margin)
            y2 = min(image.height, y2 + self.margin)
            
            # Überprüfe minimale Größe
            if (x2 - x1 >= self.min_region_size[0] and 
                y2 - y1 >= self.min_region_size[1]):
                final_regions.append((x1, y1, x2, y2))
        
        logger.info(f"Gefunden {len(final_regions)} sensitive Regionen zum Cropping")
        return final_regions

    def _find_word_boxes_for_text_span(self, 
                                      target_text: str, 
                                      word_boxes: List[Tuple[str, Tuple[int, int, int, int]]],
                                      start_pos: int = None,
                                      end_pos: int = None) -> List[Tuple[int, int, int, int]]:
        """
        Findet Word-Boxes, die einem bestimmten Text entsprechen.
        """
        matching_boxes = []
        target_words = target_text.lower().split()
        
        # Einfacher Ansatz: Suche nach aufeinanderfolgenden Words
        for i, (word, bbox) in enumerate(word_boxes):
            if word.lower() in target_words:
                matching_boxes.append(bbox)
                
                # Versuche, aufeinanderfolgende passende Words zu finden
                j = i + 1
                word_idx = target_words.index(word.lower()) + 1
                
                while (j < len(word_boxes) and 
                       word_idx < len(target_words) and
                       word_boxes[j][0].lower() == target_words[word_idx]):
                    matching_boxes.append(word_boxes[j][1])
                    j += 1
                    word_idx += 1
                
                if len(matching_boxes) >= len(target_words) / 2:  # Mindestens 50% der Words gefunden
                    break
        
        return matching_boxes

    def _find_patient_data_regions(self, 
                                  patient_info: Dict, 
                                  word_boxes: List[Tuple[str, Tuple[int, int, int, int]]],
                                  full_text: str) -> List[Tuple[int, int, int, int]]:
        """
        Findet Regionen basierend auf extrahierten Patientendaten.
        """
        regions = []
        
        if patient_info.get('patient_first_name') and patient_info.get('patient_last_name'):
            full_name = f"{patient_info['patient_first_name']} {patient_info['patient_last_name']}"
            name_boxes = self._find_word_boxes_for_text_span(full_name, word_boxes)
            if name_boxes:
                # ALT:
                # regions.append(self._create_bounding_box(name_boxes))
                regions.extend(self._create_bounding_boxes_recursive(name_boxes, axis="x"))

        # Geburtsdatum
        if patient_info.get('patient_dob'):
            dob_str = str(patient_info['patient_dob'])
            dob_boxes = self._find_word_boxes_for_text_span(dob_str, word_boxes)
            if dob_boxes:
                regions.extend(self._create_bounding_boxes_recursive(dob_boxes, axis="x"))

        # Fallnummer
        if patient_info.get('casenumber'):
            case_str = str(patient_info['casenumber'])
            case_boxes = self._find_word_boxes_for_text_span(case_str, word_boxes)
            if case_boxes:
                regions.extend(self._create_bounding_boxes_recursive(case_boxes, axis="x"))

        return regions

    def _create_bounding_box(self, boxes: List[Tuple[int, int, int, int]]) -> Tuple[int, int, int, int]:
        """
        Erstellt eine umschließende Bounding Box aus mehreren kleineren Boxes.
        """
        if not boxes:
            return (0, 0, 0, 0)
        
        x1 = min(box[0] for box in boxes)
        y1 = min(box[1] for box in boxes)
        x2 = max(box[0] + box[2] for box in boxes)  # x + width
        y2 = max(box[1] + box[3] for box in boxes)  # y + height
        
        return (x1, y1, x2, y2)

    def _merge_nearby_regions(self, regions: List[Tuple[int, int, int, int]]) -> List[Tuple[int, int, int, int]]:
        """
        Führt benachbarte Regionen zusammen, die näher als merge_distance sind.
        """
        if not regions:
            return []
        
        merged = []
        regions = sorted(regions)  # Sortiere nach x1-Koordinate
        
        current_region = regions[0]
        
        for next_region in regions[1:]:
            # Prüfe, ob Regionen nah genug sind, um sie zusammenzuführen
            if self._regions_should_merge(current_region, next_region):
                # Führe Regionen zusammen
                current_region = (
                    min(current_region[0], next_region[0]),
                    min(current_region[1], next_region[1]),
                    max(current_region[2], next_region[2]),
                    max(current_region[3], next_region[3])
                )
            else:
                merged.append(current_region)
                current_region = next_region
        
        merged.append(current_region)
        return merged

    def _regions_should_merge(self, region1: Tuple[int, int, int, int], region2: Tuple[int, int, int, int]) -> bool:
        """
        Bestimmt, ob zwei Regionen zusammengeführt werden sollten.
        """
        x1_1, y1_1, x2_1, y2_1 = region1
        x1_2, y1_2, x2_2, y2_2 = region2
        
        # Berechne Abstand zwischen Regionen
        horizontal_distance = max(0, max(x1_1, x1_2) - min(x2_1, x2_2))
        vertical_distance = max(0, max(y1_1, y1_2) - min(y2_1, y2_2))
        
        # Führe zusammen, wenn eine der Distanzen unter dem Schwellwert liegt
        return (horizontal_distance <= self.merge_distance or 
                vertical_distance <= self.merge_distance)

    def crop_sensitive_regions(self, 
                              pdf_path: str, 
                              output_dir: str,
                              page_numbers: Optional[List[int]] = None) -> Dict[str, List[str]]:
        """
        Croppt sensitive Regionen aus einem PDF und speichert sie als separate Bilder.
        
        Args:
            pdf_path: Pfad zum PDF
            output_dir: Ausgabeverzeichnis für gecropte Bilder
            page_numbers: Spezifische Seitenzahlen zu verarbeiten (None = alle Seiten)
            
        Returns:
            Dictionary mit Seite -> Liste von Crop-Bild-Pfaden
        """
        try:
            output_dir = Path(output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)
            
            # Konvertiere PDF zu Bildern
            logger.info(f"Konvertiere PDF zu Bildern: {pdf_path}")
            images = convert_pdf_to_images(pdf_path)
            
            if page_numbers is None:
                page_numbers = list(range(len(images)))
            
            results = {}
            pdf_name = Path(pdf_path).stem
            
            for page_num in page_numbers:
                if page_num >= len(images):
                    logger.warning(f"Seite {page_num} existiert nicht in PDF")
                    continue
                    
                logger.info(f"Verarbeite Seite {page_num + 1}")
                image = images[page_num]
                
                # Führe OCR durch, um Word-Boxes zu erhalten
                full_text, word_boxes = tesseract_full_image_ocr(image)
                
                if not word_boxes:
                    logger.warning(f"Keine Textboxen auf Seite {page_num + 1} gefunden")
                    continue
                
                # Erkenne sensitive Regionen
                sensitive_regions = self.detect_sensitive_regions(image, word_boxes)
                
                if not sensitive_regions:
                    logger.info(f"Keine sensitiven Regionen auf Seite {page_num + 1} gefunden")
                    results[f"page_{page_num + 1}"] = []
                    continue
                
                # Croppe und speichere sensitive Regionen
                page_crops = []
                for i, (x1, y1, x2, y2) in enumerate(sensitive_regions):
                    # Croppe die Region
                    cropped_image = image.crop((x1, y1, x2, y2))
                    
                    # Erstelle Dateinamen
                    crop_filename = f"{pdf_name}_page_{page_num + 1}_region_{i + 1}.png"
                    crop_path = output_dir / crop_filename
                    
                    # Speichere das gecropte Bild
                    cropped_image.save(crop_path)
                    page_crops.append(str(crop_path))
                    
                    logger.info(f"Gespeichert: {crop_filename} ({x2-x1}x{y2-y1} px)")
                
                results[f"page_{page_num + 1}"] = page_crops
        except Exception as e:
            logger.error(f"Fehler beim Cropping von Seite {page_num + 1}: {e}")
            results[f"page_{page_num + 1}"] = []
        
        return results

    def visualize_sensitive_regions(self, 
                                   image: Image.Image, 
                                   word_boxes: List[Tuple[str, Tuple[int, int, int, int]]],
                                   output_path: str) -> None:
        """
        Visualisiert erkannte sensitive Regionen auf dem Bild für Debugging.
        
        Args:
            image: PIL Image
            word_boxes: OCR Word-Boxes
            output_path: Pfad für das Ausgabebild
        """
        # Kopiere das Bild für Visualisierung
        vis_image = image.copy()
        draw = ImageDraw.Draw(vis_image)
        
        # Erkenne sensitive Regionen
        sensitive_regions = self.detect_sensitive_regions(image, word_boxes)
        
        # Zeichne sensitive Regionen in Rot
        for i, (x1, y1, x2, y2) in enumerate(sensitive_regions):
            draw.rectangle([x1, y1, x2, y2], outline="red", width=3)
            draw.text((x1, y1-20), f"Region {i+1}", fill="red")
        
        # Zeichne alle OCR Word-Boxes in Blau (optional)
        for word, (x, y, w, h) in word_boxes:
            draw.rectangle([x, y, x+w, y+h], outline="blue", width=1)
        
        vis_image.save(output_path)
        logger.info(f"Visualisierung gespeichert: {output_path}")

    def create_anonymized_pdf_with_crops(self, 
                                        pdf_path: str, 
                                        crop_output_dir: str,
                                        anonymized_pdf_path: str) -> None:
        """
        Erstellt eine anonymisierte Version des PDFs, bei der sensitive Regionen
        durch schwarze Rechtecke überdeckt werden.
        
        Args:
            pdf_path: Ursprüngliches PDF
            crop_output_dir: Verzeichnis mit den Crop-Daten
            anonymized_pdf_path: Pfad für das anonymisierte PDF
        """
        try:
            import fitz  # PyMuPDF
            
            logger.info(f"Erstelle anonymisiertes PDF: {anonymized_pdf_path}")
            
            # Öffne das ursprüngliche PDF
            doc = fitz.open(pdf_path)
            
            # Konvertiere PDF zu Bildern für die Analyse
            images = convert_pdf_to_images(pdf_path)
            
            # Verarbeite jede Seite
            for page_num, image in enumerate(images):
                page = doc[page_num]
                
                # Führe OCR durch, um sensitive Bereiche zu finden
                full_text, word_boxes = tesseract_full_image_ocr(image)
                
                # Erkenne sensitive Regionen
                sensitive_regions = self.detect_sensitive_regions(image, word_boxes)
                
                if sensitive_regions:
                    logger.info(f"Anonymisiere {len(sensitive_regions)} Bereiche auf Seite {page_num + 1}")
                    
                    # Konvertiere Pixel-Koordinaten zu PDF-Koordinaten
                    # PDF-Koordinaten haben Ursprung unten links, Bilder oben links
                    page_rect = page.rect
                    page_height = page_rect.height
                    page_width = page_rect.width
                    
                    img_width, img_height = image.size
                    
                    # Skalierungsfaktoren
                    scale_x = page_width / img_width
                    scale_y = page_height / img_height
                    
                    for x1, y1, x2, y2 in sensitive_regions:
                        # Konvertiere Bildkoordinaten zu PDF-Koordinaten
                        pdf_x1 = x1 * scale_x
                        pdf_y1 = page_height - (y2 * scale_y)  # Y-Achse umkehren
                        pdf_x2 = x2 * scale_x
                        pdf_y2 = page_height - (y1 * scale_y)  # Y-Achse umkehren
                        
                        # Erstelle schwarzes Rechteck
                        rect = fitz.Rect(pdf_x1, pdf_y1, pdf_x2, pdf_y2)
                        
                        # Füge schwarzes Rechteck hinzu
                        page.draw_rect(rect, color=(0, 0, 0), fill=(0, 0, 0))
                        
                        logger.debug(f"Geschwärzt: ({pdf_x1:.1f}, {pdf_y1:.1f}, {pdf_x2:.1f}, {pdf_y2:.1f})")
            
            # Speichere das anonymisierte PDF
            doc.save(anonymized_pdf_path)
            doc.close()
            
            logger.info(f"Anonymisiertes PDF gespeichert: {anonymized_pdf_path}")
            
        except ImportError:
            logger.error("PymuPDF not installed.")
            raise
        except Exception as e:
            logger.error(f"Fehler beim Erstellen des anonymisierten PDFs: {e}")
            raise


def crop_sensitive_regions_from_pdf(pdf_path: str, 
                                   output_dir: str, 
                                   margin: int = 20,
                                   visualize: bool = False) -> Dict[str, List[str]]:
    """
    Convenience-Funktion zum Cropping sensitiver Regionen aus einem PDF.
    
    Args:
        pdf_path: Pfad zum PDF
        output_dir: Ausgabeverzeichnis
        margin: Zusätzlicher Rand um sensitive Bereiche
        visualize: Ob Visualisierungen erstellt werden sollen
        
    Returns:
        Dictionary mit gecropten Bildpfaden pro Seite
    """
    cropper = SensitiveRegionCropper(margin=margin)
    
    results = cropper.crop_sensitive_regions(pdf_path, output_dir)
    
    if visualize:
        # Erstelle Visualisierungen für Debugging
        vis_dir = Path(output_dir) / "visualizations"
        vis_dir.mkdir(exist_ok=True)
        
        images = convert_pdf_to_images(pdf_path)
        pdf_name = Path(pdf_path).stem
        
        for page_num, image in enumerate(images):
            full_text, word_boxes = tesseract_full_image_ocr(image)
            vis_path = vis_dir / f"{pdf_name}_page_{page_num + 1}_visualization.png"
            cropper.visualize_sensitive_regions(image, word_boxes, str(vis_path))
    
    return results



from django import forms
from endoreg_db.models.medical import Examination

class ExaminationForm(forms.ModelForm):
    class Meta:
        model = Examination
        fields = '__all__'
        widgets = {
            'date': forms.DateInput(attrs={'type': 'date', 'class': 'form-control'}),
            'time': forms.TimeInput(attrs={'type': 'time', 'class': 'form-control'}),
        }



"""
Centralized environment configuration for EndoReg-DB.

This module is the single place to read environment variables and .env files.
It avoids loading .env during pytest, and provides typed helpers.
No Django imports here to prevent early settings configuration.
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Dict, Optional

# Detect pytest early to avoid loading .env in test runs
IS_PYTEST = bool(os.environ.get("PYTEST_CURRENT_TEST"))

# Compute repository BASE_DIR (repo root). This file is endoreg_db/config/env.py.
BASE_DIR = Path(__file__).resolve().parents[2]

# Optional: load .env only when not under pytest
_DOTENV_LOADED = False
try:
    if not IS_PYTEST:
        import dotenv

        dotenv.load_dotenv()
        _DOTENV_LOADED = True
except Exception:
    # dotenv is optional, ignore errors
    _DOTENV_LOADED = False


def _get(key: str, default: Optional[str] = None) -> Optional[str]:
    return os.environ.get(key, default)


def env_str(key: str, default: str = "") -> str:
    val = _get(key)
    return val if val is not None else default


def env_bool(key: str, default: bool = False) -> bool:
    val = _get(key)
    if val is None:
        return default
    return str(val).strip().lower() in {"1", "true", "yes", "on"}


def env_int(key: str, default: int = 0) -> int:
    val = _get(key)
    if val is None:
        return default
    try:
        return int(str(val).strip())
    except Exception:
        return default


def env_path(key: str, default_relative: str) -> Path:
    """Return an absolute path. If env is relative, resolve under BASE_DIR."""
    val = _get(key)
    if not val:
        p = BASE_DIR / default_relative
    else:
        p = Path(val)
        if not p.is_absolute():
            p = (BASE_DIR / p).resolve()
    return p


def snapshot() -> Dict[str, Any]:
    """Return a snapshot of relevant config for debugging/logging."""
    keys = [
        # Core
        "DJANGO_SETTINGS_MODULE",
        "TIME_ZONE",
        # Paths
        "STORAGE_DIR",
        "ASSET_DIR",
        "STATIC_URL",
        "MEDIA_URL",
        # Dev DB
        "DEV_DB_ENGINE",
        "DEV_DB_NAME",
        # Test DB
        "TEST_DB_ENGINE",
        "TEST_DB_NAME",
        "TEST_DB_FILE",
        # Flags
        "RUN_VIDEO_TESTS",
        "SKIP_EXPENSIVE_TESTS",
    ]
    data: Dict[str, Any] = {k: os.environ.get(k) for k in keys}
    data.update(
        {
            "IS_PYTEST": IS_PYTEST,
            "DOTENV_LOADED": _DOTENV_LOADED,
            "BASE_DIR": str(BASE_DIR),
        }
    )
    return data


# Back-compat short aliases used by settings modules
ENV = os.environ.get



from django.urls import include, path

# Import the package-level URL patterns (defined in endoreg_db/urls/__init__.py)
from endoreg_db.urls import urlpatterns as api_urlpatterns

# Expose API under /api/
urlpatterns = [
    path("api/", include((api_urlpatterns, "endoreg_db"), namespace="api")),
]



"""Module for utility classes and functions."""

# --- Imports from submodules ---

# dataloader
from endoreg_db.utils.video.ffmpeg_wrapper import assemble_video_from_frames, get_stream_info, transcode_video, transcode_videofile_if_required

from .dataloader import load_model_data_from_yaml

# dates
from .dates import ensure_aware_datetime, random_day_by_month_year, random_day_by_year

# env
from .env import DEBUG, DJANGO_SETTINGS_MODULE, get_env_var

# file_operations
from .file_operations import copy_with_progress, get_uuid_filename, rename_file_uuid

# hashs
from .hashs import (
    DJANGO_NAME_SALT,
    get_examiner_hash,
    get_hash_string,
    get_patient_examination_hash,
    get_pdf_hash,
    get_video_hash,
)

# names
from .names import (
    create_mock_examiner_name,
    create_mock_patient_name,
    guess_name_gender,
)

# parse_and_generate_yaml
from .parse_and_generate_yaml import collect_center_names

# paths
from .paths import data_paths

# pydantic_models
from .pydantic_models import DbConfig
from .storage import (
    delete_field_file,
    ensure_local_file,
    save_local_file,
)
from .storage import (
    file_exists as storage_file_exists,
)

# validate_endo_roi
from .validate_endo_roi import validate_endo_roi
from .video import split_video

# ffmpeg_wrapper
from .video.ffmpeg_wrapper import (
    extract_frames,
)

# --- Exports ---

__all__ = [
    "collect_center_names",
    "copy_with_progress",
    "create_mock_examiner_name",
    "create_mock_patient_name",
    "data_paths",
    "DbConfig",
    "DEBUG",
    "DJANGO_NAME_SALT",
    "DJANGO_SETTINGS_MODULE",
    "ensure_aware_datetime",
    "get_env_var",
    "get_examiner_hash",
    "get_hash_string",
    "get_patient_examination_hash",
    "get_pdf_hash",
    "get_uuid_filename",
    "get_video_hash",
    "guess_name_gender",
    "load_model_data_from_yaml",
    "random_day_by_month_year",
    "random_day_by_year",
    "rename_file_uuid",
    "validate_endo_roi",
    "assemble_video_from_frames",  # Updated name
    "get_stream_info",
    "transcode_video",
    "transcode_videofile_if_required",  # Added
    "extract_frames",  # Added
    "split_video",
    "delete_field_file",
    "ensure_local_file",
    "save_local_file",
    "storage_file_exists",
]
